import json

import pytest

from ..common_models import KeywordSet, Molecule
from ..gridoptimization import GridOptimizationRecord
from ..records import ResultRecord, OptimizationRecord
from ..torsiondrive import TorsionDriveRecord

## Molecule hashes


def test_molecule_water_canary_hash():

    water_dimer_minima = Molecule.from_data("""
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    --
    O   1.350625   0.111469   0.000000
    H   1.680398  -0.373741  -0.758561
    H   1.680398  -0.373741   0.758561
    """,
                                            dtype="psi4")
    assert water_dimer_minima.get_hash() == "42f3ac52af52cf2105c252031334a2ad92aa911c"

    # Check orientation
    mol = water_dimer_minima.orient_molecule()
    assert mol.get_hash() == "632490a0601500bfc677e9277275f82fbc45affe"

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "d0b499739f763e8d3a5556b4ddaeded6a148e4d5"
    assert frag_1.get_hash() == "bdc1f75bd1b7b999ff24783d7c1673452b91beb9"

@pytest.mark.parametrize("geom", [
    [0, 0, 0, 0, 5, 0],
    [0, 0, 0, 0, 5, 0 + 1.e-12],
    [0, 0, 0, 0, 5, 0 - 1.e-12],
    [0, 0, 0, 0, 5, 0 + 1.e-7],
]) # yapf: disable
def test_molecule_geometry_canary_hash(geom):

    mol = Molecule(geometry=geom, symbols=["H", "H"])

    assert mol.get_hash() == "fb69e6744407b220a96d6ddab4ec2099619db791"


## Keyword Set hash

@pytest.mark.parametrize("data, hash_index", [

    # Simple checks
    ({"values": {"hi": 5}},
     "53042da4ac1af059816008631e4589a0cd0c98c6"),

    ({"values": {"HI": 5}},
     "53042da4ac1af059816008631e4589a0cd0c98c6"),

    # Check options combinations
    ({"values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}},
     "888837dec981e2f9a1ef2ef2d33db6d239df65da"),

    ({"values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}, "lowercase": False},
     "08f6f37fec721b753096d71f7dbbccb734e8a696"),

    ({"values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}, "exact_floats": True},
     "691c86202af20868bba457514e63c83d0444142e"),

    ({"values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}, "exact_floats": True, "lowercase": False},
     "a3a6eea9edec3c2901a76fa8a41d4f9db9b3405f"),

    # Check recursive
    ({"values": {"d1": {"D2": [1.e-17, 5]}}},
     "55c6f1c7e610ce379feaa8a97854ea61974e0d92"),

    ({"values": {"d1": {"D2": [1.e-17, 5]}}, "exact_floats": True, "lowercase": False},
     "639d207f6a58ed2974f9737c3240ebffdf459857"),

    # Check hash_index build
    ({"values": {}, "hash_index": "waffles"},
     "waffles"),

    ({"values": {}, "hash_index": "waffles", "build_index": True},
     "bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f"),
]) # yapf: disable
def test_keywords_canary_hash(data, hash_index):
    """
    Canary test aginst possible changes in the options hash_index
    """

    opt = KeywordSet(**data)

    assert hash_index == opt.hash_index, opt.values


@pytest.mark.parametrize("data1, data2", [
    # Test case sensitivity
    ({"values": {}},
     {"values": {}}),

    # Test float creation
    ({"values": {"CAPS": 5, "other": 4.e-3}},
     {"values": {"caps": 5, "other": 0.004}}),

    ({"values": {"other": 4.123e-5}},
     {"values": {"OTHER": 0.00004123}}),

    # Test list of floats
    ({"values": {"other": [1.11e-2, 2.22e-3]}},
     {"values": {"other": [0.0111, 0.00222]}}),

    # Test small floats
    ({"values": {"other": 1 + 1.e-17}},
     {"values": {"other": 1.0 - 1.e-17}}),

    # Test dict order
    ({"values": {"a": 5, "b": 6, "c": None}},
     {"values": {"b": 6, "a": 5, "c": None}}),

    # Check recusive
    ({"values": {"d1": {"D2": [0.0, 5], "d3": (3, 1.e-17)}}},
     {"values": {"d1": {"d2": [1.e-17, 5], "d3": (3, 0)}}}),

]) # yapf: disable
def test_keywords_comparison_hash(data1, data2):
    """
    Ensure the hash_index finds collisions correctly before and after serialization.
    """
    opt1 = KeywordSet(**data1)
    opt2 = KeywordSet(**data2)

    # Check after serialization and rebuild index
    opt1s = KeywordSet(**json.loads(opt1.json()), build_index=True)
    opt2s = KeywordSet(**json.loads(opt2.json()), build_index=True)

    # Paranoid, try all combinations
    assert opt1.hash_index == opt2.hash_index
    assert opt1.hash_index == opt2s.hash_index
    assert opt1s.hash_index == opt2.hash_index
    assert opt1s.hash_index == opt2s.hash_index


@pytest.mark.parametrize("model", [ResultRecord, OptimizationRecord])
def test_hash_fields(model):
    assert "procedure" in model.get_hash_fields()
    assert "program" in model.get_hash_fields()


## Record hashes
_base_result = {
    "driver": "gradient",
    "method": "hf",
    "basis": "sto-3g",
    "keywords": None,
    "program": "prog",
    "molecule": "5c7896fb95d592ad07a2fe3b"
}
@pytest.mark.parametrize("data, hash_index", [

    # Check same
    ({}, "e1e20d5c13c8ad7ba894f71af39f3c0884ef2aca"),

    ({"program": "PROG"},
     "e1e20d5c13c8ad7ba894f71af39f3c0884ef2aca"),

    ({"method": "HF"},
     "e1e20d5c13c8ad7ba894f71af39f3c0884ef2aca"),

    # All different
    ({"program": "prog2"},
     "f8638510976e5146b6bef79995f6f14acd06f645"),

    ({"driver": "energy"},
     "4c2c737694f9438c73b294d7f72249acbbd3b649"),

    ({"keywords": "5c7896fb95d592ad07a2fe3b"},
     "32e115a991ffb85ac0c29bf4dd5d93e733919b5c"),

    # Check same
    ({"basis": ""},
     "1ae5df953fc38f9b866bd50af6d5513dbfc49cc6"),

    ({"basis": "null"},
     "1ae5df953fc38f9b866bd50af6d5513dbfc49cc6"),

    ({"basis": None},
     "1ae5df953fc38f9b866bd50af6d5513dbfc49cc6"),

]) # yapf: disable

def test_result_record_canary_hash(data, hash_index):

    opt = ResultRecord(**{**_base_result, **data})

    assert hash_index == opt.get_hash_index(), data
    assert opt.hash_index is None  # Not set


## Optimization hashes
_qc_spec = {"driver": "gradient", "method": "HF", "basis": "sto-3g", "keywords": None, "program": "prog"}
_base_opt = {"keywords": {}, "program": "prog2", "initial_molecule": "5c7896fb95d592ad07a2fe3b", "qc_spec": _qc_spec}
@pytest.mark.parametrize("data, hash_index", [

    # Check same
    ({}, "254de59f1598570d0c31aa2d3d84b601c9da12b9"),

    ({"program": "PROG2"},
     "254de59f1598570d0c31aa2d3d84b601c9da12b9"),

    ({"qc_spec": {**_qc_spec, **{"program": "prog"}}},
     "254de59f1598570d0c31aa2d3d84b601c9da12b9"),

    ({"qc_spec": {**_qc_spec, **{"method": "HF"}}},
     "254de59f1598570d0c31aa2d3d84b601c9da12b9"),

    # Check tolerances
    ({"keywords": {"tol": 1.e-12}},
     "8ab52bff9430f7759323e6a547afc58725422c47"),

    ({"keywords": {"tol": 0.0}},
     "8ab52bff9430f7759323e6a547afc58725422c47"),

    ({"keywords": {"tol": 1.e-9}},
     "1628caf9a29c9bf17a66cb55b13106e7f2704e51"), # Should be different from above

    # Check fields
    ({"initial_molecule": "5c78987e95d592ad07a2fe3c"},
     "3c20c8f7b1be857460f2d71d74680dd19e9d9113"),

    # Check basis preps
    ({"qc_spec": {**_qc_spec, **{"basis": None}}},
     "3489e0c47144ebedb4fdcc2bfab61f7aa4dc947c"),

    ({"qc_spec": {**_qc_spec, **{"basis": ""}}},
     "3489e0c47144ebedb4fdcc2bfab61f7aa4dc947c"),
]) # yapf: disable

def test_optimization_record_canary_hash(data, hash_index):

    opt = OptimizationRecord(**{**_base_opt, **data})

    assert hash_index == opt.hash_index, data


## GridOptimizationRecord hashes
_opt_spec = {
    "program": "geometric",
    "keywords": {
        "coordsys": "tric",
    }
}

_scan_spec = {
        "type": "distance",
        "indices": [1, 2],
        "steps": [-0.1, 0.0],
        "step_type": "relative"
    } # yapf: disable

_base_gridopt = {
    "keywords": {
        "preoptimization": False,
        "scans": [_scan_spec]
    },
    "optimization_spec": _opt_spec,
    "qc_spec": _qc_spec,
    "initial_molecule": "5c7896fb95d592ad07a2fe3b",
    "starting_molecule": "5c7896fb95d592ad07a2fe3b",
    "grid_optimizations": {},
    "final_energy_dict": {},
    "starting_grid": tuple(),
    "provenance": {
        "creator": ""
    }
}


@pytest.mark.parametrize("data, hash_index", [

    # Check same
    ({}, "fa2da83aae1651a9115f5eaea83043187c4c8c7b"),
    ({"keywords": {
            "preoptimization": False,
            "scans": [{**_scan_spec, **{"steps": [-0.1 + 1e-12, 0.0 - 1.e-12]}}]
        }
    }, "fa2da83aae1651a9115f5eaea83043187c4c8c7b"),

    # Check opt keywords stability
    ({"optimization_spec": {**_opt_spec, **{"keywords": {"tol": 1.e-12}}}
    }, "fa2da83aae1651a9115f5eaea83043187c4c8c7b"),

    ({"optimization_spec": {**_opt_spec, **{"keywords": {"tol": 0}}}
    }, "fa2da83aae1651a9115f5eaea83043187c4c8c7b"),

    # Check fields
    ({"initial_molecule": "5c78987e95d592ad07a2fe3c"},
     "9624ce2eca96eabcdb9ec3b2e073429f6dd4b8a4"),

    ({"qc_spec": {**_qc_spec, **{"method": "custom1"}}
    },"d6a187bf6de7a25d36402c5d16109ddc6f4f217d"),
]) # yapf: disable
def test_gridoptimization_canary_hash(data, hash_index):

    gridopt = GridOptimizationRecord(**{**_base_gridopt, **data})

    assert hash_index == gridopt.get_hash_index(), data


## TorsionDriveRecord hashes

_base_torsion = {
    "keywords": {
        "dihedrals": [[0, 1, 2, 3]],
        "grid_spacing": [10]
    },
    "optimization_spec": _opt_spec,
    "qc_spec": _qc_spec,
    "initial_molecule": ["5c7896fb95d592ad07a2fe3b"],
    "final_energy_dict": {},
    "optimization_history": {},
    "minimum_positions": {},
    "provenance": {
        "creator": ""
    }
}


@pytest.mark.parametrize(
    "data, hash_index",
    [
        ({}, "dd305011ee2b741b1dcd03350994920a3718b289"),

        # Check same
        ({
            "keywords": {
                "dihedrals": [[0, 1, 2, 3]],
                "grid_spacing": [10],
                "tol": 1.e-12
            }
        }, "cb3f9c9bd4eda742b0429ebea0c3d12719ab2582"),
        ({
            "keywords": {
                "dihedrals": [[0, 1, 2, 3]],
                "grid_spacing": [10],
                "tol": 0
            }
        }, "cb3f9c9bd4eda742b0429ebea0c3d12719ab2582"),
        ({
            "keywords": {
                "dihedrals": [[0, 1, 2, 3]],
                "grid_spacing": [10],
                "tol": 1.e-9
            }
        }, "903cc0deb4f0e7b8bc41a69cf5fbd0c9420176a4"),

        # Check opt keywords stability
        ({
            "optimization_spec": {
                **_opt_spec,
                **{
                    "keywords": {
                        "tol": 0.0
                    }
                }
            }
        }, "a12fd524b0e215b5252b464ca4041091916df8bb"),
        ({
            "optimization_spec": {
                **_opt_spec,
                **{
                    "keywords": {
                        "tol": 1.e-12
                    }
                }
            }
        }, "a12fd524b0e215b5252b464ca4041091916df8bb"),

        # Check fields
        ({
            "initial_molecule": ["5c78987e95d592ad07a2fe3c"]
        }, "f209751a4a6559a8d2d539c070f3b701d1ddf9f2"),
    ])  # yapf disable
def test_torsiondrive_canary_hash(data, hash_index):

    td = TorsionDriveRecord(**{**_base_torsion, **data})

    assert hash_index == td.get_hash_index(), data



"""
Misc information and runtime information.
"""

from importlib.util import find_spec

from . import _version

__all__ = ["get_information"]

versions = _version.get_versions()

__info = {"version": versions['version'], "git_revision": versions['full-revisionid']}


def _isnotebook():
    """
    Checks if we are inside a jupyter notebook or not.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell in ['ZMQInteractiveShell', 'google.colab._shell']:
            return True
        elif shell == 'TerminalInteractiveShell':
            return False
        else:
            return False
    except NameError:
        return False


__info["isnotebook"] = _isnotebook()


def find_module(name):
    return find_spec(name)


def get_information(key):
    """
    Obtains a variety of runtime information about QCFractal.
    """
    key = key.lower()
    if key not in __info:
        raise KeyError("Information key '{}' not understood.".format(key))

    return __info[key]


def provenance_stamp(routine):
    """Return dictionary satisfying QCSchema,
   generating routine's name is passed in through `routine`.

   """
    return {
        'creator': 'QCFractal',
        'version': get_information('version'),
        'routine': routine,
    }


__all__ = ['parsl_dict', 'base_dict']

base_dict = {
    "MANAGER_NAME_SHORT": "Parsl",
    "MANAGER_NAME_LONG": "Parsl Parallel Scripting Library",
    "MANAGER_URL_TITLE": "Parsl",
    "MANAGER_URL": "https://parsl.readthedocs.io/en/latest/index.html",
    "MANAGER_CLIENT": "parsl_config",
    "SANITY_CHECKS": "",
    "IMPORTS": """
from parsl.config import Config
from parsl.executors.ipp import HighThroughputExecutor
""",

    "MANAGER_CLIENT_BUILDER": ""
}

code_skeletal = """
config = Config(
    executors=[
        HighThroughputExecutor(
            label='QCFractal Compute Executor',
            provider={PROVIDER}(
                queue=TORQUE_QUEUE,
                account=TORQUE_ACCOUNT,
                scheduler_options=SCHEDULER_OPTS,
                worker_init=TASK_STARTUP_COMMANDS,
                walltime="00:10:00",
                init_blocks=1,
                max_blocks=MAX_NODES,
                nodes_per_block=1,        # Keep one node per block, its just easier this way
            ),
            workers_per_node=MAX_TASKS_PER_NODE,
        )

    ],
)

"""

# SLURM

slurm_imports = base_dict["IMPORTS"] + "\n" + """"from parsl.providers import TorqueProvider"""

slurm_sanity_checks = """
if NODE_EXCLUSIVITY and "--exclusive" not in SCHEDULER_OPTS:
    SCHEDULER_OPTS.append("--exclusive")
"""

slurm_builder = """
cluster = SLURMCluster(
    name='QCFractal Dask Compute Executor',
    cores=CORES_PER_NODE,
    memory=str(MEMORY_PER_NODE) + "GB",
    queue=SLURM_PARTITION,
    processes=MAX_TASKS_PER_NODE,  # This subdivides the cores by the number of processes we expect to run
    walltime="00:10:00",

    # Additional queue submission flags to set
    job_extra=SCHEDULER_OPTS,
    # Not sure of the validity of this, but it seems to be the only terminal-invoking way
    # so python envs may be setup from there
    # Commands to execute before the Dask
    env_extra=TASK_STARTUP_COMMANDS
)
"""

# PBS/Torque

torque_imports = base_dict["IMPORTS"] + "\n" + """"from dask_jobqueue import PBSCluster"""

torque_builder = """
cluster = PBSCluster(
    name='QCFractal Dask Compute Executor',
    cores=CORES_PER_NODE,
    memory=str(MEMORY_PER_NODE) + "GB",
    queue=TORQUE_QUEUE,
    project=TORQUE_ACCOUNT,
    processes=MAX_TASKS_PER_NODE,  # This subdivides the cores by the number of processes we expect to run
    walltime="00:10:00",

    # Additional queue submission flags to set
    job_extra=SCHEDULER_OPTS,
    # Not sure of the validity of this, but it seems to be the only terminal-invoking way
    # so python envs may be setup from there
    # Commands to execute before the Dask
    env_extra=TASK_STARTUP_COMMANDS
)
"""

# LSF

lsf_imports = base_dict["IMPORTS"] + "\n" + """"from dask_jobqueue import LSFCluster"""

lsf_builder = """
cluster = LSFCluster(
    name='QCFractal Dask Compute Executor',
    cores=CORES_PER_NODE,
    memory=str(MEMORY_PER_NODE) + "GB",
    queue=LSF_QUEUE,
    project=LSF_PROJECT,
    processes=MAX_TASKS_PER_NODE,  # This subdivides the cores by the number of processes we expect to run
    walltime="00:10:00",

    # Additional queue submission flags to set
    job_extra=SCHEDULER_OPTS,
    # Not sure of the validity of this, but it seems to be the only terminal-invoking way
    # so python envs may be setup from there
    # Commands to execute before the Dask
    env_extra=TASK_STARTUP_COMMANDS
)
"""


# Final

parsl_dict = {
    "slurm": {**base_dict,
              **{"SANITY_CHECKS": slurm_sanity_checks,
                 "IMPORTS": slurm_imports,
                 "MANAGER_CLIENT_BUILDER": code_skeletal.format(BUILDER=slurm_builder),
                 }
              },
    "torque": {**base_dict,
               **{"IMPORTS": torque_imports,
                  "MANAGER_CLIENT_BUILDER": code_skeletal.format(BUILDER=torque_builder)}
               },
    "lsf": {**base_dict,
            **{"IMPORTS": lsf_imports,
               "MANAGER_CLIENT_BUILDER": code_skeletal.format(BUILDER=lsf_builder)}
            }
}




try:
    import flask
    import dash
    import dash_core_components
    import dash_html_components
except ModuleNotFound:
    raise ModuleNotFound("Could not find `dash`, please `conda install dash -c conda-forge` to use the dashboard.")

try:
    import dash_bootstrap_components
except ModuleNotFound:
    raise ModuleNotFound("Could not find `dash-bootstrap-components`, please `conda install dash-bootstrap-components -c conda-forge` to use the dashboard.")

from . import index
from .app import app as dashboard_app



"""
Mongoengine Database class to handle access to mongoDB through ODM.
"""

try:
    import pymongo
except ImportError:
    raise ImportError(
        "Mongostorage_socket requires pymongo, please install this python module or try a different db_socket.")

try:
    import mongoengine
except ImportError:
    raise ImportError(
        "Mongoengine_socket requires mongoengine, please install this python module or try a different db_socket.")

import logging
import secrets
from datetime import datetime as dt
from typing import Any, Dict, List, Optional, Union

import bcrypt
import bson.errors
import mongoengine as db
import mongoengine.errors
from bson.objectid import ObjectId
from mongoengine.connection import disconnect, get_db

from .me_models import (CollectionORM, KeywordsORM, KVStoreORM, MoleculeORM, ProcedureORM, QueueManagerORM, ResultORM,
                        ServiceQueueORM, TaskQueueORM, UserORM)
from .storage_utils import add_metadata_template, get_metadata_template
from ..interface.models import KeywordSet, Molecule, ResultRecord, TaskRecord, prepare_basis


def _str_to_indices_with_errors(ids: List[Union[str, ObjectId]]):
    if isinstance(ids, (str, ObjectId)):
        ids = [ids]

    good = []
    bad = []
    for x in ids:
        if isinstance(x, str):
            try:
                good.append(ObjectId(x))
            except bson.errors.InvalidId:
                bad.append(x)
        elif isinstance(x, ObjectId):
            good.append(x)
        else:
            bad.append(x)
    return good, bad


_null_keys = {"basis", "keywords"}
_id_keys = {"id", "molecule", "keywords", "procedure_id"}
_lower_func = lambda x: x.lower()
_upper_func = lambda x: x.upper()
_prepare_keys = {"program": _lower_func, "basis": prepare_basis, "method": _lower_func, "procedure": _lower_func, "status": _upper_func}


def format_query(**query: Dict[str, Union[str, List[str]]]) -> Dict[str, Union[str, List[str]]]:
    """
    Formats a query into a MongoEngine description.
    """

    ret = {}
    errors = []
    for k, v in query.items():
        if v is None:
            continue

        # Handle None keys
        k = k.lower()
        if (k in _null_keys) and (v == 'null'):
            v = None

        # Handle ID conversions
        elif k in _id_keys:
            if isinstance(v, (list, tuple)):
                v, bad = _str_to_indices_with_errors(v)
                if bad:
                    errors.append((k, bad))
            else:
                v = ObjectId(v)

        if k in _prepare_keys:
            f = _prepare_keys[k]
            if isinstance(v, (list, tuple)):
                v = [f(x) for x in v]
            else:
                v = f(v)

        if isinstance(v, (list, tuple)):
            ret[k + "__in"] = v
        else:
            ret[k] = v

    return ret, errors


class MongoengineSocket:
    """
        Mongoengine QCDB wrapper class.
    """

    def __init__(self,
                 uri: str,
                 project: str="molssidb",
                 bypass_security: bool=False,
                 allow_read: bool=True,
                 authMechanism: str="SCRAM-SHA-1",
                 authSource: str=None,
                 logger: 'Logger'=None,
                 max_limit: int=1000):
        """
        Constructs a new socket where url and port point towards a Mongod instance.

        """

        # Logging data
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('MongoengineSocket')

        # Security
        self._bypass_security = bypass_security
        self._allow_read = allow_read

        self._lower_results_index = ["method", "basis", "program"]

        # disconnect from any active default connection
        disconnect()

        # Build MongoClient
        expanded_uri = pymongo.uri_parser.parse_uri(uri)
        if expanded_uri["password"] is not None:
            # connect to mongoengine
            self.client = db.connect(db=project, host=uri, authMechanism=authMechanism, authSource=authSource)
        else:
            # connect to mongoengine
            self.client = db.connect(db=project, host=uri)

        self._url, self._port = expanded_uri["nodelist"][0]

        try:
            version_array = self.client.server_info()['versionArray']

            if tuple(version_array) < (3, 2):
                raise RuntimeError
        except AttributeError:
            raise RuntimeError(
                "Could not detect MongoDB version at URL {}. It may be a very old version or installed incorrectly. "
                "Choosing to stop instead of assuming version is at least 3.2.".format(uri))
        except RuntimeError:
            # Trap low version
            raise RuntimeError("Connected MongoDB at URL {} needs to be at least version 3.2, found version {}.".
                               format(uri, self.client.server_info()['version']))

        # Isolate objects to this single project DB
        self._project_name = project
        self._tables = self.client[project]
        self._max_limit = max_limit

    ### Mongo meta functions

    def __str__(self) -> str:
        return "<MongoSocket: address='{0:s}:{1:d}:{2:s}'>".format(str(self._url), self._port, str(self._project_name))

    def _clear_db(self, db_name: str):
        """Dangerous, make sure you are deleting the right DB."""

        self.logger.warning("Clearing database '{}' and dropping all tables.".format(db_name))

        # make sure it's the right DB
        if get_db().name == db_name:
            self.logger.info('Clearing database: {}'.format(db_name))
            ResultORM.drop_collection()
            MoleculeORM.drop_collection()
            KeywordsORM.drop_collection()
            KVStoreORM.drop_collection()
            CollectionORM.drop_collection()
            TaskQueueORM.drop_collection()
            ServiceQueueORM.drop_collection()
            QueueManagerORM.drop_collection()
            ProcedureORM.drop_collection()
            UserORM.drop_collection()

            self.client.drop_database(db_name)

        QueueManagerORM.objects(name='').first()  # init

    def _delete_DB_data(self, db_name):
        self._clear_db(db_name)

    def get_project_name(self) -> str:
        return self._project_name

    def get_limit(self, limit: Optional[int]) -> int:
        """Get the allowed limit on results to return in queries based on the
         given `limit`. If this number is greater than the
         mongoengine_soket.max_limit then the max_limit will be returned instead.
        """

        return limit if limit and limit < self._max_limit else self._max_limit

### KV Functions

    def add_kvstore(self, blobs_list: List[Any]):
        """
        Adds to the key/value store table.

        Parameters
        ----------
        blobs_list : List[Any]
            A list of data blobs to add.

        Returns
        -------
        A dict with keys: 'data' and 'meta'
            (see get_metadata_template())
            The 'data' part is an object of the result or None if not found
        """

        meta = add_metadata_template()
        blob_ids = []
        for blob in blobs_list:
            if blob is None:
                blob_ids.append(None)
                continue

            doc = KVStoreORM(value=blob)
            doc.save()
            blob_ids.append(str(doc.id))
            meta['n_inserted'] += 1

        meta["success"] = True

        return {"data": blob_ids, "meta": meta}

    def get_kvstore(self, id: List[str]=None,  limit: int=None, skip: int=0):
        """
        Pulls from the key/value store table.

        Parameters
        ----------
        id : List[str]
            A list of ids to query

        Returns
        -------
        A dict with keys: 'data' and 'meta'
            (see get_metadata_template())
            The 'data' part is an object of the result or None if not found
        """

        meta = get_metadata_template()

        query, errors = format_query(id=id)

        data = KVStoreORM.objects(**query).limit(self.get_limit(limit))\
                                          .skip(skip)

        meta["success"] = True
        meta["n_found"] = data.count()  # all data count, can be > len(data)
        meta["errors"].extend(errors)

        data = [d.to_json_obj() for d in data]
        data = {d["id"]: d["value"] for d in data}
        return {"data": data, "meta": meta}

### Molecule functions

    def get_add_molecules_mixed(self, data: List[Union[str, Molecule]]) -> List[Molecule]:
        """
        Get or add the given molecules (if they don't exit).
        MoleculeORMs are given in a mixed format, either as a dict of mol data
        or as existing mol id.

        TODO: to be split into get by_id and get_by_data
        """

        meta = get_metadata_template()

        ordered_mol_dict = {indx: mol for indx, mol in enumerate(data)}
        new_molecules = {}
        id_mols = {}
        for idx, mol in ordered_mol_dict.items():
            if isinstance(mol, str):
                id_mols[idx] = mol
            elif isinstance(mol, Molecule):
                new_molecules[idx] = mol
            else:
                meta["errors"].append((idx, "Data type not understood"))

        ret_mols = {}

        # Add all new molecules
        flat_mols = []
        flat_mol_keys = []
        for k, v in new_molecules.items():
            flat_mol_keys.append(k)
            flat_mols.append(v)
        flat_mols = self.add_molecules(flat_mols)["data"]
        id_mols.update({k: v for k, v in zip(flat_mol_keys, flat_mols)})

        # Get molecules by index and translate back to dict
        tmp = self.get_molecules(list(id_mols.values()))
        id_mols_list = tmp["data"]
        meta["errors"].extend(tmp["meta"]["errors"])

        inv_id_mols = {v: k for k, v in id_mols.items()}

        for mol in id_mols_list:
            ret_mols[inv_id_mols[mol.id]] = mol

        meta["success"] = True
        meta["n_found"] = len(ret_mols)
        meta["missing"] = list(ordered_mol_dict.keys() - ret_mols.keys())

        # Rewind to flat last
        ret = []
        for ind in range(len(ordered_mol_dict)):
            if ind in ret_mols:
                ret.append(ret_mols[ind])
            else:
                ret.append(None)

        return {"meta": meta, "data": ret}

    def add_molecules(self, molecules: List[Molecule]):
        """
        Adds molecules to the database.

        Parameters
        ----------
        data : dict of molecule-like JSON objects
            A {key: molecule} dictionary of molecules to input.

        Returns
        -------
        bool
            Whether the operation was successful.
        """

        meta = add_metadata_template()

        results = []
        for dmol in molecules:

            mol_dict = dmol.json_dict(exclude={"id"})

            mol_dict["fix_com"] = True
            mol_dict["fix_orientation"] = True

            # Build fresh indices
            mol_dict["molecule_hash"] = dmol.get_hash()
            mol_dict["molecular_formula"] = dmol.get_molecular_formula()

            mol_dict["identifiers"] = {}
            mol_dict["identifiers"]["molecule_hash"] = mol_dict["molecule_hash"]
            mol_dict["identifiers"]["molecular_formula"] = mol_dict["molecular_formula"]

            # search by index keywords not by all keys, much faster
            doc = MoleculeORM.objects(molecule_hash=mol_dict['molecule_hash'])

            if doc.count() == 0:
                doc = MoleculeORM(**mol_dict).save()
                results.append(str(doc.id))
                meta['n_inserted'] += 1
            else:

                id = str(doc.first().id)
                meta['duplicates'].append(id)  # TODO
                # If new or duplicate, add the id to the return list
                results.append(id)

                # We should make sure there was not a hash collision?
                # new_mol.compare(old_mol)
                # raise KeyError("!!! WARNING !!!: Hash collision detected")
        meta["success"] = True

        ret = {"data": results, "meta": meta}
        return ret

    def get_molecules(self, id=None, molecule_hash=None, molecular_formula=None, limit: int=None, skip: int=0):

        ret = {"meta": get_metadata_template(), "data": []}

        query, errors = format_query(id=id, molecule_hash=molecule_hash, molecular_formula=molecular_formula)

        # Don't include the hash or the molecular_formula in the returned result
        # Make the query
        data = MoleculeORM.objects(**query).exclude("molecule_hash", "molecular_formula")\
                                        .limit(self.get_limit(limit))\
                                        .skip(skip)\

        ret["meta"]["success"] = True
        ret["meta"]["n_found"] = data.count()  # all data count, can be > len(data)
        ret["meta"]["errors"].extend(errors)

        # Data validated going in
        data = [Molecule(**d.to_json_obj(), validate=False) for d in data]
        ret["data"] = data

        return ret

    def del_molecules(self, id: List[str]=None, molecule_hash: List[str]=None):
        """
        Removes a molecule from the database from its hash.

        Parameters
        ----------
        values : str or list of strs
            The hash of a molecule.

        Returns
        -------
        int
            Number of deleted molecules.
        """

        query, errors = format_query(id=id, molecule_hash=molecule_hash)
        return MoleculeORM.objects(**query).delete()

    ### Mongo options functions

    def add_keywords(self, keyword_sets: List[KeywordSet]):
        """Add one KeywordSet uniquly identified by 'program' and the 'name'.

        Parameters
        ----------
         data
            A list of KeywordSets to be inserted.

        Returns
        -------
            A dict with keys: 'data' and 'meta'
            (see add_metadata_template())
            The 'data' part is a list of ids of the inserted options
            data['duplicates'] has the duplicate entries

        Notes
        ------
            Duplicates are not considered errors.

        """

        meta = add_metadata_template()

        keywords = []
        for kw in keyword_sets:

            kw_dict = kw.json_dict(exclude={"id"})

            # search by index keywords not by all keys, much faster
            found = KeywordsORM.objects(hash_index=kw_dict['hash_index']).first()
            if not found:

                doc = KeywordsORM(**kw_dict).save()
                keywords.append(str(doc.id))
                meta['n_inserted'] += 1
            else:
                meta['duplicates'].append(str(found.id))  # TODO
                keywords.append(str(found.id))
            meta["success"] = True

        ret = {"data": keywords, "meta": meta}
        return ret

    def get_keywords(self,
                     id: Union[str, list]=None,
                     hash_index: Union[str, list]=None,
                     limit: int=None,
                     skip: int=0,
                     return_json: bool=True,
                     with_ids: bool=True) -> List[KeywordSet]:
        """Search for one (unique) option based on the 'program'
        and the 'name'. No overwrite allowed.

        Parameters
        ----------
        id : list or str
            Ids of the keywords
        hash_index : list or str
            hash index of keywords
        limit : int, optional
            Maximum number of results to return.
            If this number is greater than the mongoengine_soket.max_limit then
            the max_limit will be returned instead.
            Default is to return the socket's max_limit (when limit=None or 0)
        skip : int, optional
        return_json : bool, optional
            Return the results as a json object
            Default is True
        with_ids : bool, optional
            Include the DB ids in the returned object (names 'id')
            Default is True


        Returns
        -------
            A dict with keys: 'data' and 'meta'
            (see get_metadata_template())
            The 'data' part is an object of the result or None if not found
        """

        meta = get_metadata_template()
        query, errors = format_query(id=id, hash_index=hash_index)

        data = []
        try:
            data = KeywordsORM.objects(**query).limit(self.get_limit(limit)).skip(skip)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:  # TODO: remove
            meta['error_description'] = str(err)

        if return_json:
            rdata = [KeywordSet(**d.to_json_obj(with_ids)) for d in data]
        else:
            rdata = data

        return {"data": rdata, "meta": meta}

    def get_add_keywords_mixed(self, data):
        """
        Get or add the given options (if they don't exit).
        KeywordsORM are given in a mixed format, either as a dict of mol data
        or as existing mol id.

        TODO: to be split into get by_id and get_by_data
        """

        meta = get_metadata_template()

        ids = []
        for idx, kw in enumerate(data):
            if isinstance(kw, str):
                ids.append(kw)

            elif isinstance(kw, KeywordSet):
                new_id = self.add_keywords([kw])["data"][0]
                ids.append(new_id)
            else:
                meta["errors"].append((idx, "Data type not understood"))
                ids.append(None)

        missing = []
        ret = []
        for idx, id in enumerate(ids):
            if id is None:
                ret.append(None)
                missing.append(idx)
                continue

            tmp = self.get_keywords(id=id)["data"]
            if tmp:
                ret.append(tmp[0])
            else:
                ret.append(None)

        meta["success"] = True
        meta["n_found"] = len(ret) - len(missing)
        meta["missing"] = missing

        return {"meta": meta, "data": ret}

    def del_keywords(self, id: str) -> int:
        """
        Removes a option set from the database based on its keys.

        Parameters
        ----------
        program : str
            The program of the option set
        name : str
            The name of the option set

        Returns
        -------
        int
           number of deleted documents
        """

        # monogoengine
        count = 0
        kw = KeywordsORM.objects(id=ObjectId(id))
        if kw:
            count = kw.delete()

        return count

    ### Mongo database functions

    # def add_collection(self, data, overwrite=False):
    def add_collection(self, data: Dict[str, Any], overwrite: bool=False):
        """Add (or update) a collection to the database.

        Parameters
        ----------
        collection : str
        name : str
        data : dict
        overwrite : bool
            Update existing collection

        Returns
        -------
        A dict with keys: 'data' and 'meta'
            (see add_metadata_template())
            The 'data' part is the id of the inserted document or none

        Notes
        -----
        ** Change: The data doesn't have to include the ID, the document
        is identified by the (collection, name) pairs.
        ** Change: New fields will be added to the collection, but existing won't
            be removed.
        """

        meta = add_metadata_template()
        col_id = None
        try:

            if ("id" in data) and (data["id"] == "local"):
                data.pop("id", None)
            lname = data.get("name").lower()
            collection = data.pop("collection").lower()

            if overwrite:
                # may use upsert=True to add or update
                col = CollectionORM.objects(collection=collection, lname=lname).update_one(**data)
            else:
                col = CollectionORM(collection=collection, lname=lname, **data).save()

            meta['success'] = True
            meta['n_inserted'] = 1
            col_id = str(col.id)
        except Exception as err:
            meta['error_description'] = str(err)

        ret = {'data': col_id, 'meta': meta}
        return ret

    # def get_collections(self, keys, projection=None):
    def get_collections(self,
                        collection: str=None,
                        name: str=None,
                        return_json: bool=True,
                        with_ids: bool=True,
                        limit: int=None,
                        projection: Dict[str, Any]=None,
                        skip: int=0) -> Dict[str, Any]:
        """Get collection by collection and/or name

        Parameters
        ----------
        collection : str, optional
        name : str, optional
        return_json : bool
        with_ids : bool
        limit : int
        skip : int

        Returns
        -------
        A dict with keys: 'data' and 'meta'
            The data is a list of the collections found
        """

        meta = get_metadata_template()
        if name:
            name = name.lower()
        if collection:
            collection = collection.lower()
        query, errors = format_query(lname=name, collection=collection)

        data = []
        try:
            if projection:
                data = CollectionORM.objects(**query).only(*projection).limit(self.get_limit(limit)).skip(skip)
            else:
                data = CollectionORM.objects(**query).exclude("lname").limit(self.get_limit(limit)).skip(skip)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            rdata = [d.to_json_obj(with_ids) for d in data]
        else:
            rdata = data

        return {"data": rdata, "meta": meta}

    def del_collection(self, collection: str, name: str) -> bool:
        """
        Remove a collection from the database based on its keys.

        Parameters
        ----------
        collection: str
            CollectionORM type
        name : str
            CollectionORM name

        Returns
        -------
        int
            Number of documents deleted
        """
        return CollectionORM.objects(collection=collection.lower(), lname=name.lower()).delete()

## ResultORMs functions

    def add_results(self, record_list: List[ResultRecord]):
        """
        Add results from a given dict. The dict should have all the required
        keys of a result.

        Parameters
        ----------
        data : list of dict
            Each dict must have:
            program, driver, method, basis, options, molecule
            Where molecule is the molecule ID in the DB
            In addition, it should have the other attributes that it needs
            to store

        Returns
        -------
        Dict with keys: data, meta
            Data is the ids of the inserted/updated/existing docs
        """

        meta = add_metadata_template()

        result_ids = []
        # try:
        for result in record_list:

            doc = ResultORM.objects(
                program=result.program,
                driver=result.driver,
                method=result.method,
                basis=result.basis,
                keywords=result.keywords,
                molecule=result.molecule)

            if doc.count() == 0:
                doc = ResultORM(**result.json_dict(exclude={"id"})).save()
                result_ids.append(str(doc.id))
                meta['n_inserted'] += 1
            else:
                id = str(doc.first().id)
                meta['duplicates'].append(id)  # TODO
                # If new or duplicate, add the id to the return list
                result_ids.append(id)
        meta["success"] = True

        ret = {"data": result_ids, "meta": meta}
        return ret

    def update_results(self, record_list: List[ResultRecord]):
        """
        Update results from a given dict (replace existing).

        Parameters
        ----------
        id : list of str
            IDs of the results to update, must exist in the DB
        data : list of dict
            Data that needs to be updated
            Shouldn't update:
            program, driver, method, basis, options, molecule

        Returns
        -------
            number of records updated
        """

        # try:
        updated_count = 0
        for result in record_list:

            if result.id is None:
                logger.error("Attempted update without ID, skipping")
                continue

            ResultORM(**result.json_dict()).save()
            updated_count += 1

        return updated_count

    def get_results_count(self):
        """
        TODO: just return the count, used for big queries.

        Returns
        -------

        """
        pass

    def get_results(self,
                    id: Union[str, List]=None,
                    program: str=None,
                    method: str=None,
                    basis: str=None,
                    molecule: str=None,
                    driver: str=None,
                    keywords: str=None,
                    task_id: Union[str, List]=None,
                    status: str='COMPLETE',
                    projection=None,
                    limit: int=None,
                    skip: int=0,
                    return_json=True,
                    with_ids=True):
        """

        Parameters
        ----------
        id : str or list
        program : str
        method : str
        basis : str
        molecule : str
            MoleculeORM id in the DB
        driver : str
        keywords : str
            The id of the option in the DB
        task_id : List of str or str
            Task id that ran the results
        status : bool, default is 'COMPLETE'
            The status of the result: 'COMPLETE', 'INCOMPLETE', or 'ERROR'
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is 0
            skip the first 'skip' results. Used to paginate
        return_json : bool, default is True
            Return the results as a list of json instead of objects
        with_ids : bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = get_metadata_template()

        # Ignore status if Id or task_id is present
        if id is not None or task_id is not None:
            status = None

        query, error = format_query(
            id=id,
            program=program,
            method=method,
            basis=basis,
            molecule=molecule,
            driver=driver,
            keywords=keywords,
            status=status)

        q_limit = self.get_limit(limit)

        data = []
        try:
            if projection:
                data = ResultORM.objects(**query).only(*projection).limit(q_limit).skip(skip)
            else:
                data = ResultORM.objects(**query).limit(q_limit).skip(skip)

            meta["n_found"] = data.count()  # total number found, can be >len(data)
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def del_results(self, ids: List[str]):
        """
        Removes results from the database using their ids
        (Should be cautious! other tables maybe referencing results).

        Parameters
        ----------
        ids : list of str
            The Ids of the results to be deleted

        Returns
        -------
        int
            number of results deleted
        """

        obj_ids = [ObjectId(x) for x in ids]

        return ResultORM.objects(id__in=obj_ids).delete()

### Mongo procedure/service functions

    def add_procedures(self, record_list: List['BaseRecord']):
        """
        Add procedures from a given dict. The dict should have all the required
        keys of a result.

        Parameters
        ----------
        data : list of dict
            Each dict must have:
            procedure, program, keywords, qc_meta, hash_index
            In addition, it should have the other attributes that it needs
            to store
        update_existing : bool (default False)
            Update existing results

        Returns
        -------
        Dict with keys: data, meta
            Data is the ids of the inserted/updated/existing docs
        """

        meta = add_metadata_template()

        procedure_ids = []
        for procedure in record_list:
            doc = ProcedureORM.objects(hash_index=procedure.hash_index)

            if doc.count() == 0:
                doc = doc.upsert_one(**procedure.json_dict(exclude={"id"}))
                procedure_ids.append(str(doc.id))
                meta['n_inserted'] += 1
            else:
                id = str(doc.first().id)
                meta['duplicates'].append(id)  # TODO
                procedure_ids.append(id)
        meta["success"] = True

        ret = {"data": procedure_ids, "meta": meta}
        return ret

    def get_procedures(self,
                       id: Union[str, List]=None,
                       procedure: str=None,
                       program: str=None,
                       hash_index: str=None,
                       task_id: Union[str, List]=None,
                       status: str='COMPLETE',
                       projection=None,
                       limit: int=None,
                       skip: int=0,
                       return_json=True,
                       with_ids=True):
        """

        Parameters
        ----------
        id : str or list
        procedure : str
        program : str
        hash_index : str
        task_id : str or list
        status : bool, default is 'COMPLETE'
            The status of the result: 'COMPLETE', 'INCOMPLETE', or 'ERROR'
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is 0
            skip the first 'skip' results. Used to paginate
        return_json : bool, default is True
            Return the results as a list of json instead of objects
        with_ids : bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = get_metadata_template()

        if id is not None or task_id is not None:
            status = None

        query, error = format_query(
            id=id, procedure=procedure, program=program, hash_index=hash_index, task_id=task_id, status=status)

        q_limit = self.get_limit(limit)

        data = []
        try:
            if projection:
                data = ProcedureORM.objects(**query).only(*projection).limit(q_limit).skip(skip)
            else:
                data = ProcedureORM.objects(**query).limit(q_limit).skip(skip)

            meta["n_found"] = data.count()  # all data count
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def update_procedures(self, records_list: List['BaseRecord']):
        """
        TODO: to be updated with needed
        """

        updated_count = 0
        for procedure in records_list:

            # Must have ID
            if procedure.id is None:
                self.logger.error(
                    "No procedure id found on update (hash_index={}), skipping.".format(procedure.hash_index))
                continue

            ProcedureORM(**procedure.json_dict()).save()
            updated_count += 1

        return updated_count

    def add_services(self, service_list: List['BaseService']):
        """
        Add services from a given dict.

        Parameters
        ----------
        data : list of dict

        Returns
        -------
        Dict with keys: data, meta
            Data is the hash_index of the inserted/existing docs
        """

        meta = add_metadata_template()

        procedure_ids = []
        for service in service_list:

            # Add the underlying procedure
            new_procedure = self.add_procedures([service.output])

            # ProcedureORM already exists
            proc_id = new_procedure["data"][0]
            if new_procedure["meta"]["duplicates"]:
                procedure_ids.append(proc_id)
                meta["duplicates"].append(proc_id)
                continue

            # search by hash index
            doc = ServiceQueueORM.objects(hash_index=service.hash_index)
            service.procedure_id = proc_id

            if doc.count() == 0:
                doc = ServiceQueueORM(**service.json_dict(exclude={"id"}))
                doc.save()
                procedure_ids.append(proc_id)
                meta['n_inserted'] += 1
            else:
                procedure_ids.append(None)
                meta["errors"].append((idx, "Duplicate service, but not caught by procedure."))

        meta["success"] = True

        ret = {"data": procedure_ids, "meta": meta}
        return ret

    def get_services(self,
                     id: Union[List[str], str]=None,
                     procedure_id: Union[List[str], str]=None,
                     hash_index: Union[List[str], str]=None,
                     status: str=None,
                     projection=None,
                     limit: int=None,
                     skip: int=0,
                     return_json=True):
        """

        Parameters
        ----------
        id / hash_index : List of str or str
            service id / hash_index that ran the results
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is 0
            skip the first 'skip' results. Used to paginate
        return_json : bool, deafult is True
            Return the results as a list of json instead of objects

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = get_metadata_template()
        query, error = format_query(id=id, hash_index=hash_index, procedure_id=procedure_id, status=status)

        q_limit = self.get_limit(limit)

        data = []
        # try:
        if projection:
            services = ServiceQueueORM.objects(**query).only(*projection).limit(q_limit).skip(skip)
        else:
            services = ServiceQueueORM.objects(**query).limit(q_limit).skip(skip)

        meta["n_found"] = services.count()
        meta["success"] = True
        # except Exception as err:
        #     meta['error_description'] = str(err)

        if services.count() and return_json:
            data = [d.to_json_obj() for d in services]

        return {"data": data, "meta": meta}

    def update_services(self, records_list: List["BaseService"]) -> int:
        """
        Replace existing service.

        Raises exception if the id is invalid.

        Parameters
        ----------
        id
        updates

        Returns
        -------
            if operation is succesful
        """

        updated_count = 0
        for service in records_list:
            if service.id is None:
                self.logger.error(
                    "No service id found on update (hash_index={}), skipping.".format(service.hash_index))
                continue

            ServiceQueueORM(**service.json_dict()).save()
            updated_count += 1

        return updated_count

    def services_completed(self, records_list: List["BaseService"]) -> int:

        done = 0
        for service in records_list:
            if service.id is None:
                self.logger.error(
                    "No service id found on completion (hash_index={}), skipping.".format(service.hash_index))
                continue

            procedure = service.output
            procedure.id = service.procedure_id
            self.update_procedures([procedure])

            ServiceQueueORM.objects(id=ObjectId(service.id)).delete()

            done += 1

        return done

### Mongo queue handling functions

    def queue_submit(self, data: List[TaskRecord]):
        """Submit a list of tasks to the queue.
        Tasks are unique by their base_result, which should be inserted into
        the DB first before submitting it's corresponding task to the queue
        (with result.status='INCOMPLETE' as the default).
        The default task.status is 'WAITING'.

        Duplicate tasks sould be a rare case.
        Hooks are merged if the task already exists.

        Parameters
        ----------
        data : list of tasks (dict)
            A task is a dict, with the following fields:
            - hash_index: idx, not used anymore
            - spec: dynamic field (dict-like), can have any structure
            - tag: str
            - base_results: tuple (required), first value is the class type
             of the result, {'results' or 'procedure'). The second value is
             the ID of the result in the DB. Example:
             "base_result": ('results', result_id)

        Returns
        -------
        dict (data and meta)
            'data' is a list of the IDs of the tasks IN ORDER, including
            duplicates. An errored task has 'None' in its ID
            meta['duplicates'] has the duplicate tasks
        """

        meta = add_metadata_template()

        results = []
        for task_num, record in enumerate(data):
            try:

                result_obj = None
                if record.base_result.ref == 'result':
                    result_obj = ResultORM(id=record.base_result.id)
                elif record.base_result.ref == 'procedure':
                    result_obj = ProcedureORM(id=record.base_result.id)
                else:
                    raise TypeError("Base_result type must be 'results' or 'procedure',"
                                    " {} is given.".format(record.base_result.ref))
                task = TaskQueueORM(**record.json_dict(exclude={"id", "base_result"}))
                task.base_result = result_obj
                task.save()

                result_obj.update(task_id=str(task.id))  # update bidirectional rel
                results.append(str(task.id))
                meta['n_inserted'] += 1
            except mongoengine.errors.NotUniqueError as err:  # rare case
                # If results is stored as DBRef, get it with:
                # task = TaskQueueORM.objects(__raw__={'base_result': base_result}).first()  # avoid

                # If base_result is stored as a ResultORM or ProcedureORM class, get it with:
                task = TaskQueueORM.objects(base_result=result_obj).first()
                self.logger.warning('queue_submit got a duplicate task: {}'.format(task.to_mongo()))
                results.append(str(task.id))
                meta['duplicates'].append(task_num)
            except Exception as err:
                self.logger.warning('queue_submit submission error: {}'.format(str(err)))
                meta["success"] = False
                meta["errors"].append(str(err))
                results.append(None)

        meta["success"] = True

        ret = {"data": results, "meta": meta}
        return ret

    def queue_get_next(self, manager, available_programs, available_procedures, limit=100, tag=None,
                       as_json=True) -> List[TaskRecord]:
        """TODO: needs to be done in a transaction."""

        # Figure out query, tagless has no requirements
        query, error = format_query(
            status="WAITING",
            program=available_programs,
            procedure=available_procedures,  # Procedure can be none, explicitly include
            tag=tag)
        query["procedure__in"].append(None)

        found = TaskQueueORM.objects(**query).limit(limit).order_by('-priority', 'created_on')

        query = {"_id": {"$in": [x.id for x in found]}}

        # update_many using pymongo in one DB access
        upd = TaskQueueORM._get_collection().update_many(
            query, {"$set": {
                "status": "RUNNING",
                "modified_on": dt.utcnow(),
                "manager": manager,
            }})

        if as_json:
            found = [TaskRecord(**task.to_json_obj()) for task in found]

        if upd.modified_count != len(found):
            self.logger.warning("QUEUE: Number of found projects does not match the number of updated projects.")

        return found

    def get_queue(self,
                  id=None,
                  hash_index=None,
                  program=None,
                  status: str=None,
                  base_result: str=None,
                  projection=None,
                  limit: int=None,
                  skip: int=0,
                  return_json=False,
                  with_ids=True):
        """
        TODO: check what query keys are needed.
        Parameters
        ----------
        id : list
            Ids of the tasks
        Hash_index
        status : bool, default is None (find all)
            The status of the task: 'COMPLETE', 'RUNNING', 'WAITING', or 'ERROR'
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is None 0
            skip the first 'skip' results. Used to paginate
        return_json : bool, default is True
            Return the results as a list of json instead of objects
        with_ids : bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = get_metadata_template()
        query, error = format_query(program=program, id=id, hash_index=hash_index, status=status)

        if base_result:
            query['__raw__'] = {'base_result._ref.$id': ObjectId(base_result)}
        q_limit = self.get_limit(limit)

        data = []
        try:
            if projection:
                data = TaskQueueORM.objects(**query).only(*projection).limit(q_limit).skip(skip)
            else:
                data = TaskQueueORM.objects(**query).limit(q_limit).skip(skip)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)


        data = [TaskRecord(**task.to_json_obj()) for task in data]

        return {"data": data, "meta": meta}

    def queue_get_by_id(self, id: List[str], limit: int=None, skip: int=0, as_json: bool=True):
        """Get tasks by their IDs.

        Parameters
        ----------
        id : list of str
            List of the task IDs in the DB
        limit : int (optional)
            max number of returned tasks. If limit > max_limit, max_limit
            will be returned instead (safe query)
        as_json : bool
            Return tasks as JSON

        Returns
        -------
        list of the found tasks
        """

        found = TaskQueueORM.objects(id__in=id).limit(self.get_limit(limit)).skip(skip)

        if as_json:
            found = [TaskRecord(**task.to_json_obj()) for task in found]

        return found

    def queue_mark_complete(self, task_ids: List[str]) -> int:
        """Update the given tasks as complete.
        Note that each task is already pointing to its result location.
        Mark the corresponding result/procedure as complete.

        Parameters
        ----------
        task_ids : list
            IDs of the tasks to mark as COMPLETE

        Returns
        -------
        int
            Updated count
        """

        # If using replica sets, we can use as a transcation:
        # with self.client.start_session() as session:
        #     with session.start_transaction():
        #         # automatically calls ClientSession.commit_transaction().
        #         # If the block exits with an exception, the transaction
        #         # automatically calls ClientSession.abort_transaction().
        #           found = TaskQueueORM._get_collection().update_many(
        #                               {'id': {'$in': task_ids}},
        #                               {'$set': {'status': 'COMPLETE'}},
        #                               session=session)
        #         # next, Update results

        tasks = TaskQueueORM.objects(id__in=task_ids).update(status='COMPLETE', modified_on=dt.utcnow())
        results = ResultORM.objects(task_id__in=task_ids).update(status='COMPLETE', modified_on=dt.utcnow())
        procedures = ProcedureORM.objects(task_id__in=task_ids).update(status='COMPLETE', modified_on=dt.utcnow())

        # This should not happen unless there is data inconsistency in the DB
        if results + procedures < tasks:
            self.logger.error("Some tasks don't reference results or procedures correctly!"
                              "Tasks: {}, ResultORMs: {}, procedures: {}. ".format(tasks, results, procedures))
        return tasks

    def queue_mark_error(self, data):
        """update the given tasks as errored.
        Mark the corresponding result/procedure as complete.

        """

        bulk_commands = []
        bulk_commands_records = []
        task_ids = []
        for task_id, msg in data:
            task_ids.append(task_id)
            update = {
                "$set": {
                    "status": "ERROR",
                    "error": msg,
                    "modified_on": dt.utcnow(),
                }
            }
            bulk_commands.append(pymongo.UpdateOne({"_id": ObjectId(task_id)}, update))

            # Update the objects as well, different from mark complete as these are not processed the same way
            # This design should be overhauled...
            error_id = self.add_kvstore([msg])["data"][0]
            update = {
                "$set": {
                    "status": "ERROR",
                    "error": error_id,
                    "modified_on": dt.utcnow(),
                }
            }
            # Task id is held as a string here... don't move to ObjectId
            bulk_commands_records.append(pymongo.UpdateOne({"task_id": task_id}, update))

        if len(bulk_commands) == 0:
            return

        task_mod = TaskQueueORM._get_collection().bulk_write(bulk_commands, ordered=False).modified_count
        rec_mod = ResultORM._get_collection().bulk_write(bulk_commands_records, ordered=False).modified_count
        rec_mod += ProcedureORM._get_collection().bulk_write(bulk_commands_records, ordered=False).modified_count
        if task_mod != rec_mod:
            self.logger.error(
                "Queue Mark Error: Number of tasks updates {}, does not match the number of records updates {}.".
                format(task_mod, rec_mod))

        return task_mod

    def queue_reset_status(self, manager: str, reset_running: bool=True, reset_error: bool=False) -> int:
        """
        Reset the status of the tasks that a manager owns from Running to Waiting.
        If reset_error is True, then also reset errored tasks AND its results/proc.

        Parameters
        ----------
        manager : str
            The manager name to reset the status of
        reset_running : str (optional), default is True
            If True, reset running tasks to be waiting
        reset_error : str (optional), default is False
            If True, also reset errored tasks to be waiting,
            also update results/proc to be INCOMPLETE

        Returns
        -------
        int
            Updated count
        """

        if not (reset_running or reset_error):
            # nothing to do
            return 0

        # Update results and procedures if reset_error
        if reset_error:
            task_objs = TaskQueueORM.objects(manager=manager, status="ERROR").only('id')
            task_ids = [x.id for x in task_objs]
            ResultORM.objects(task_id__in=task_ids).update(status='INCOMPLETE', modified_on=dt.utcnow())
            ProcedureORM.objects(task_id__in=task_ids).update(status='INCOMPLETE', modified_on=dt.utcnow())

        status = []
        if reset_running:
            status.append("RUNNING")
        if reset_error:
            status.append("ERROR")

        updated = TaskQueueORM.objects(
            manager=manager, status__in=status).update(
                status="WAITING", modified_on=dt.utcnow())

        return updated

    def del_tasks(self, id: Union[str, list]):
        """Delete a task from the queue. Use with caution.

        Parameters
        ----------
        id : str or list
            Ids of the tasks to delete
        Returns
        -------
        int
            Number of tasks deleted
        """

        return TaskQueueORM.objects(id__in=id).delete()

### QueueManagerORMs

    def manager_update(self, name, **kwargs):

        inc_count = {
            # Increment relevant data
            "inc__submitted": kwargs.pop("submitted", 0),
            "inc__completed": kwargs.pop("completed", 0),
            "inc__returned": kwargs.pop("returned", 0),
            "inc__failures": kwargs.pop("failures", 0)
        }

        upd = {key: kwargs[key] for key in QueueManagerORM._fields_ordered if key in kwargs}

        QueueManagerORM.objects()  # init
        manager = QueueManagerORM.objects(name=name)
        if manager:  # existing
            upd.update(inc_count)
            num_updated = manager.update(**upd, modified_on=dt.utcnow())
        else:  # create new, ensures defaults and validations
            QueueManagerORM(name=name, **upd).save()
            num_updated = 1

        return num_updated == 1

    def get_managers(self, name: str=None, status: str=None, modified_before=None):

        query, error = format_query(name=name, status=status)
        if modified_before:
            query["modified_on__lt"] = modified_before

        data = QueueManagerORM.objects(**query)

        meta = get_metadata_template()
        meta["success"] = True
        meta["n_found"] = data.count()

        data = [x.to_json_obj(with_id=False) for x in data]

        return {"data": data, "meta": meta}

### UserORMs

    def add_user(self,
                 username: str,
                 password: Optional[str]=None,
                 permissions: List[str]=["read"],
                 *,
                 overwrite: bool=False) -> Union[bool, str]:
        """
        Adds a new user and associated permissions.

        Passwords are stored using bcrypt.

        Parameters
        ----------
        username : str
            New user's username
        password : str
            The user's password
        permissions : list of str, optional
            The associated permissions of a user ['read', 'write', 'compute', 'queue', 'admin']

        Returns
        -------
        tuple
            Successful insert or not
        """
        valid_permissions = {'read', 'write', 'compute', 'queue', 'admin'}

        # Make sure permissions are valid
        if not valid_permissions >= set(permissions):
            raise KeyError("Permissions settings not understood: {}".format(set(permissions) - valid_permissions))

        return_password = False
        if password is None:
            password = secrets.token_urlsafe(32)
            return_password = True

        hashed = bcrypt.hashpw(password.encode("UTF-8"), bcrypt.gensalt(6))
        blob = {"username": username, "password": hashed, "permissions": permissions}

        success = False
        if overwrite:
            doc = UserORM.objects(username=username)
            doc.upsert_one(**blob)
            success = True

        else:
            try:
                UserORM(**blob).save()
                success = True
            except mongoengine.errors.NotUniqueError:
                success = False

        if return_password and success:
            return password
        else:
            return success

    def verify_user(self, username, password, permission):
        """
        Verifies if a user has the requested permissions or not.

        Passwords are store and verified using bcrypt.

        Parameters
        ----------
        username : str
            The username to verify
        password : str
            The password associated with the username
        permission : str
            The associated permissions of a user ['read', 'write', 'compute', 'queue', 'admin']

        Returns
        -------
        tuple
            A tuple of (success flag, failure string)

        Examples
        --------

        >>> db.add_user("george", "shortpw")

        >>> db.verify_user("george", "shortpw", "read")
        True

        >>> db.verify_user("george", "shortpw", "admin")
        False

        """

        if self._bypass_security or (self._allow_read and (permission == "read")):
            return (True, "Success")

        data = UserORM.objects(username=username).first()
        if data is None:
            return (False, "User not found.")

        pwcheck = bcrypt.checkpw(password.encode("UTF-8"), data.password)
        if pwcheck is False:
            return (False, "Incorrect password.")

        # Admin has access to everything
        if (permission.lower() not in data.permissions) and ("admin" not in data.permissions):
            return (False, "User has insufficient permissions.")

        return (True, "Success")

    def remove_user(self, username):
        """Removes a user from the MongoDB Tables.

        Parameters
        ----------
        username : str
            The username to remove

        Returns
        -------
        bool
            If the operation was successful or not.
        """
        return UserORM.objects(username=username).delete() == 1

    def _get_users(self):

        data = UserORM.objects()

        return [x.to_json_obj(with_id=False) for x in data]

    def get_total_count(self, className, **kwargs):

        return className.objects(**kwargs).count()


"""
Parsl Manager Helper

Conditions:
- Parsl Parallel Scripting Library
- Manager running on the head node
- SLURM manager

For additional information about the Parsl, please visit this site:
https://parsl.readthedocs.io/en/latest/index.html
"""

# Fractal Settings
# Location of the Fractal Server you are connecting to
FRACTAL_URI = "localhost:7777"

# Queue Manager Settings
# Set whether or not we are just testing the Queue Manger (no Fractal Client needed)
TEST_RUN = False


# How many cores per node you want your jobs to have access to
CORES_PER_NODE = 1
# How much memory per node (in GB) you want your jobs to have access to
MEMORY_PER_NODE = 1
# How many tasks per node you want to execute on
MAX_TASKS_PER_NODE = 1
# Maximum number of nodes to try and consume
MAX_NODES = 1
# Whether or not to claim nodes for exclusive use. We recommend you do, but that's up to you
NODE_EXCLUSIVITY = True

# Generic Cluster Settings
# ========================
# Additional commands to send to the command line (often used as "#SBATCH ..." or '#PBS' headers.)
# This is a per-node type setting, not task. Don't set memory or cpu or wall clock through this
# -- Note ---
# Different Managers interpret this slightly differently, but that should not be your concern, just treat
# each item as though it were a CLI entry and the manager block will interpret
# ------------ 
SCHEDULER_OPTS = []

# Additional commands to start each task with. E.g. Activating a conda environment
# Put each command as its own item in strings
TASK_STARTUP_COMMANDS = []

# SLURM Specific Settings
# Name of the SLURM partition to draw from
SLURM_PARTITION = ''


###################

# QCFractal import
import qcfractal
import qcfractal.interface as portal

# Make sure logging is setup correctly
import tornado.log
tornado.log.enable_pretty_logging()

from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import SlurmProvider

# Quick sanity checks

if CORES_PER_NODE < 1 or not isinstance(CORES_PER_NODE, int):
    raise ValueError("CORES_PER_NODE must be an integer of at least 1")
if MAX_TASKS_PER_NODE < 1 or not isinstance(MAX_TASKS_PER_NODE, int):
    raise ValueError("MAX_TASKS_PER_NODE must be an integer of at least 1")
if MAX_NODES < 1 or not isinstance(MAX_NODES, int):
    raise ValueError("MAX_NODES must be an integer of at least 1")
if MEMORY_PER_NODE <= 0:
    raise ValueError("MEMORY_PER_NODE must be a number > 0")


parsl_config = Config(
    executors=[
        HighThroughputExecutor(
            label='QCFractal_Compute_Executor',
            provider=SlurmProvider(
                SLURM_PARTITION,
                exclusive=NODE_EXCLUSIVITY,
                scheduler_options='#SBATCH ' + '\n#SBATCH '.join(SCHEDULER_OPTS) + '\n',
                worker_init='\n'.join(TASK_STARTUP_COMMANDS),
                walltime="00:10:00",
                init_blocks=1,
                max_blocks=MAX_NODES,
                nodes_per_block=1,        # Keep one node per block, its just easier this way
            ),
            # workers_per_node=MAX_TASKS_PER_NODE,
            cores_per_worker=CORES_PER_NODE // MAX_TASKS_PER_NODE,
            max_workers = MAX_NODES*MAX_TASKS_PER_NODE
        )

    ],
)

# Build a interface to the server
# If testing, there is no need to point to a Fractal Client and None is passed in
# In production, the client is needed
if TEST_RUN:
    fractal_client = None
else:
    fractal_client = portal.FractalClient(FRACTAL_URI, verify=False)
    


# Build a manager
manager = qcfractal.queue.QueueManager(fractal_client, parsl_config, update_frequency=0.5,
                                       cores_per_task=CORES_PER_NODE // MAX_TASKS_PER_NODE,
                                       memory_per_task=MEMORY_PER_NODE // MAX_TASKS_PER_NODE)

# Important for a calm shutdown
from qcfractal.cli.cli_utils import install_signal_handlers
install_signal_handlers(manager.loop, manager.stop)

# Start or test the loop. Swap with the .test() and .start() method respectively
if TEST_RUN:
    manager.test()
else:
    manager.start()



"""
A set of utility functions to help the Collections
"""

import math
from typing import Any, Dict, List

__registered_collections = {}


def nCr(n: int, r: int) -> int:
    """
    Compute the binomial coefficient n! / (k! * (n-k)!)

    Parameters
    ----------
    n : int
        Number of samples
    r : int
        Denominator

    Returns
    -------
    ret : int
        Value
    """
    return math.factorial(n) / math.factorial(r) / math.factorial(n - r)


def register_collection(collection: 'Collection') -> None:
    """Registers a collection for use by the factory.

    Parameters
    ----------
    collection : Collection
        The Collection class to be registered

    """

    class_name = collection.__name__.lower()
    # if class_name in __registered_collections:
    #     raise KeyError("Collection type '{}' already registered".format(class_name))
    __registered_collections[class_name] = collection


def collection_factory(data: Dict[str, Any], client: 'FractalClient'=None) -> 'Collection':
    """Creates a new Collection class from a JSON blob.

    Parameters
    ----------
    data : Dict[str, Any]
        The JSON blob to create a new class from.
    client : FractalClient, optional
        A FractalClient connected to a server

    Returns
    -------
    Collection
        A ODM of the data.

    """
    if "collection" not in data:
        raise KeyError("Attempted to create Collection from JSON, but no `collection` field found.")

    if data["collection"].lower() not in __registered_collections:
        raise KeyError("Attempted to create Collection of unknown type '{}'.".format(data["collection"]))

    return __registered_collections[data["collection"].lower()].from_json(data, client=client)


def collections_name_map() -> Dict[str, str]:
    """
    Returns a map of internal name to external Collection name.

    Returns
    -------
    Dict[str, str]
        Map of {'internal': 'user fiendly name'}
    """
    return {k: v.__name__ for k, v in __registered_collections.items()}


def list_known_collections() -> List[str]:
    """
    Returns the case sensitive list of Collection names.

    Returns
    -------
    List[str]
        A list of registered collections
    """
    return list(collection_name_map.values())


def composition_planner(program=None, method=None, basis=None, driver=None, keywords=None):
    """
    Plans out a given query into multiple pieces
    """

    base = {"program": program, "method": method, "basis": basis, "driver": driver, "keywords": keywords}

    if ("-d3" in method.lower()) and ("dftd3" != program.lower()):
        dftd3keys = {"program": "dftd3", "method": method, "basis": None, "driver": driver, "keywords": None}
        base["method"] = method.split("-D3")[0]

        return [dftd3keys, base]

    else:
        return [base]



"""
A command line interface to generate qcfractal manager scripts
"""

import argparse
from textwrap import dedent, indent

__all__ = ["main"]

# Using backslash on opening triple quotes ('''\) to indicate no new line after from that line return in final

base_script = '''\
"""
{MANAGER_NAME_SHORT} Manager Helper

Conditions:
- {MANAGER_NAME_LONG}
- Manager running on the head node
- {CLUSTER_QUEUE_NAME} manager

For additional information about the {MANAGER_URL_TITLE}, please visit this site:
{MANAGER_URL}
"""

# Fractal Settings
# Location of the Fractal Server you are connecting to
FRACTAL_URI = "localhost:7777"  # QCArchive is at: api.qcarchive.molssi.org:443
# Authentication with the Fractal Server
USERNAME = None
PASSWORD = None
VERIFY_SSL = False


# Queue Manager Settings
# Set whether or not we are just testing the Queue Manger (no Fractal Client needed)
TEST_RUN = {TEST_RUN}
# Tell the manager to only pull jobs with this tag
QUEUE_TAG = None



# How many cores per node you want your jobs to have access to
CORES_PER_NODE = 1
# How much memory per node (in GB) you want your jobs to have access to
MEMORY_PER_NODE = 1
# How many tasks per node you want to execute on
MAX_TASKS_PER_NODE = 1
# Maximum number of nodes to try and consume
MAX_NODES = 1
# Whether or not to claim nodes for exclusive use. We recommend you do, but that's up to you
NODE_EXCLUSIVITY = True

# Generic Cluster Settings
# ========================
# Additional commands to send to the command line (often used as "#SBATCH ..." or '#PBS' headers.)
# This is a per-node type setting, not task. Don't set memory or cpu or wall clock through this
# -- Note ---
# Different Managers interpret this slightly differently, but that should not be your concern, just treat
# each item as though it were a CLI entry and the manager block will interpret
# ------------
SCHEDULER_OPTS = []

# Additional commands to start each task with. E.g. Activating a conda environment
# Put each command as its own item in strings
TASK_STARTUP_COMMANDS = []

{CLUSTER_QUEUE_FLAGS}

###################

# QCFractal import
import qcfractal
import qcfractal.interface as portal

# Make sure logging is setup correctly
import tornado.log
tornado.log.enable_pretty_logging()

{IMPORTS}

# Quick sanity checks

if CORES_PER_NODE < 1 or not isinstance(CORES_PER_NODE, int):
    raise ValueError("CORES_PER_NODE must be an integer of at least 1")
if MAX_TASKS_PER_NODE < 1 or not isinstance(MAX_TASKS_PER_NODE, int):
    raise ValueError("MAX_TASKS_PER_NODE must be an integer of at least 1")
if MAX_NODES < 1 or not isinstance(MAX_NODES, int):
    raise ValueError("MAX_NODES must be an integer of at least 1")
if MEMORY_PER_NODE <= 0:
    raise ValueError("MEMORY_PER_NODE must be a number > 0")
{SANITY_CHECKS}

{MANAGER_CLIENT_BUILDER}

# Build a interface to the server
# If testing, there is no need to point to a Fractal Client and None is passed in
# In production, the client is needed
if TEST_RUN:
    fractal_client = None
else:
    fractal_client = portal.FractalClient(FRACTAL_URI, 
                                          username=USERNAME,
                                          password=PASSWORD,
                                          verify=VERIFY_SSL)



# Build a manager
manager = qcfractal.queue.QueueManager(fractal_client, {MANAGER_CLIENT}, update_frequency=0.5,
                                       cores_per_task=CORES_PER_NODE // MAX_TASKS_PER_NODE,
                                       memory_per_task=MEMORY_PER_NODE // MAX_TASKS_PER_NODE,
                                       queue_tag=QUEUE_TAG)

# Important for a calm shutdown
from qcfractal.cli.cli_utils import install_signal_handlers
install_signal_handlers(manager.loop, manager.stop)

# Start or test the loop. Swap with the .test() and .start() method respectively
if TEST_RUN:
    manager.test()
else:
    manager.start()
'''

torque_helper = {
    "CLUSTER_QUEUE_NAME": "PBS/Torque",

    "CLUSTER_QUEUE_FLAGS": dedent("""\
        # TORQUE Specific Settings
        # Name of the Torque Queue to request resources from, set `None` if you don't know (may be cluster specific)
        # Equivalent to the `#PBS -q` option
        TORQUE_QUEUE = None
        # Name of the Torque account/project to charge, set `None` if using default scheduler (may be cluster specific)
        # Equivalent to the `#PBS -A` option
        TORQUE_ACCOUNT = None
        """)
}

slurm_helper = {
    "CLUSTER_QUEUE_NAME": "SLURM",

    "CLUSTER_QUEUE_FLAGS": dedent("""\
        # SLURM Specific Settings
        # Name of the SLURM partition to draw from
        SLURM_PARTITION = ''
        """)
}

lsf_helper = {
    "CLUSTER_QUEUE_NAME": "LSF",

    "CLUSTER_QUEUE_FLAGS": dedent("""\
        # LSF Specific Settings
        # Name of the LSF Queue to request resources from, set `None` if you don't know (may be cluster specific)
        # Equivalent to the `#BSUB -q` option
        LSF_QUEUE = None
        # Name of the LSF project to charge, set `None` if using default scheduler (may be cluster specific)
        # Equivalent to the `#BSUB -P` option
        LSF_PROJECT = None
        """)
}

scheduler_collections = {
    "slurm": slurm_helper,
    "torque": torque_helper,
    "lsf": lsf_helper
}

test_helper = {
    True: {  # Tests are set, comment stuff out
        "TEST_RUN": "True",
    },
    False: {  # Production, connect to Fractal
        "TEST_RUN": "False",
    }
}


def dask_templates():

    base_dict = {
        "MANAGER_NAME_SHORT": "Dask Distributed",
        "MANAGER_NAME_LONG": "Dask Distributed and Dask Job Queue (dask_jobqueue in Conda/pip)",
        "MANAGER_URL_TITLE": "Dask Job Queue",
        "MANAGER_URL": "https://jobqueue.dask.org/en/latest/",
        "MANAGER_CLIENT": "dask_client",
        "SANITY_CHECKS": "",
        "IMPORTS": """from dask.distributed import Client""",
        "MANAGER_CLIENT_BUILDER": ""
    }

    code_skeletal = dedent("""\
        {BUILDER}

        # Setup up adaption
        # Workers are distributed down to the cores through the sub-divided processes
        # Optimization may be needed
        cluster.adapt(minimum=0, maximum=MAX_NODES)

        # Integrate cluster with client
        dask_client = Client(cluster)

        """)

    # SLURM

    slurm_imports = base_dict["IMPORTS"] + "\n" + """from dask_jobqueue import SLURMCluster"""

    slurm_sanity_checks = dedent("""\
        if NODE_EXCLUSIVITY and "--exclusive" not in SCHEDULER_OPTS:
            SCHEDULER_OPTS.append("--exclusive")
        """)

    slurm_builder = dedent("""\
        cluster = SLURMCluster(
            name='QCFractal_Dask_Compute_Executor',
            cores=CORES_PER_NODE,
            memory=str(MEMORY_PER_NODE) + "GB",
            queue=SLURM_PARTITION,
            processes=MAX_TASKS_PER_NODE,  # This subdivides the cores by the number of processes we expect to run
            walltime="00:10:00",

            # Additional queue submission flags to set
            job_extra=SCHEDULER_OPTS,
            # Not sure of the validity of this, but it seems to be the only terminal-invoking way
            # so python envs may be setup from there
            # Commands to execute before the Dask
            env_extra=TASK_STARTUP_COMMANDS,
            # Uncomment and set this if your cluster uses non-standard ethernet port names
            # for communication between the head node and your compute nodes
            # interface="eth0"
            extra=['--resources process=1'],
        )
        """)

    # PBS/Torque

    torque_imports = base_dict["IMPORTS"] + "\n" + """from dask_jobqueue import PBSCluster"""

    torque_builder = dedent("""\
        cluster = PBSCluster(
            name='QCFractal_Dask_Compute_Executor',
            cores=CORES_PER_NODE,
            memory=str(MEMORY_PER_NODE) + "GB",
            queue=TORQUE_QUEUE,
            project=TORQUE_ACCOUNT,
            processes=MAX_TASKS_PER_NODE,  # This subdivides the cores by the number of processes we expect to run
            walltime="00:10:00",

            # Additional queue submission flags to set
            job_extra=SCHEDULER_OPTS,
            # Not sure of the validity of this, but it seems to be the only terminal-invoking way
            # so python envs may be setup from there
            # Commands to execute before the Dask
            env_extra=TASK_STARTUP_COMMANDS,
            # Uncomment and set this if your cluster uses non-standard ethernet port names
            # for communication between the head node and your compute nodes
            # interface="eth0"
            extra=['--resources process=1'],
        )
        """)

    # LSF

    lsf_imports = base_dict["IMPORTS"] + "\n" + """from dask_jobqueue import LSFCluster"""

    lsf_builder = dedent("""\
        cluster = LSFCluster(
            name='QCFractal_Dask_Compute_Executor',
            cores=CORES_PER_NODE,
            memory=str(MEMORY_PER_NODE) + "GB",
            queue=LSF_QUEUE,
            project=LSF_PROJECT,
            processes=MAX_TASKS_PER_NODE,  # This subdivides the cores by the number of processes we expect to run
            walltime="00:10:00",

            # Additional queue submission flags to set
            job_extra=SCHEDULER_OPTS,
            # Not sure of the validity of this, but it seems to be the only terminal-invoking way
            # so python envs may be setup from there
            # Commands to execute before the Dask
            env_extra=TASK_STARTUP_COMMANDS,
            # Uncomment and set this if your cluster uses non-standard ethernet port names
            # for communication between the head node and your compute nodes
            # interface="eth0"
            extra=['--resources process=1'],
        )
        """)

    # Final

    dask_dict = {
        "slurm": {**base_dict,
                  **{"SANITY_CHECKS": slurm_sanity_checks,
                     "IMPORTS": slurm_imports,
                     "MANAGER_CLIENT_BUILDER": code_skeletal.format(BUILDER=slurm_builder),
                     }
                  },
        "torque": {**base_dict,
                   **{"IMPORTS": torque_imports,
                      "MANAGER_CLIENT_BUILDER": code_skeletal.format(BUILDER=torque_builder)}
                   },
        "lsf": {**base_dict,
                **{"IMPORTS": lsf_imports,
                   "MANAGER_CLIENT_BUILDER": code_skeletal.format(BUILDER=lsf_builder)}
                }
    }

    return dask_dict, base_dict


def parsl_templates():

    base_dict = {
        "MANAGER_NAME_SHORT": "Parsl",
        "MANAGER_NAME_LONG": "Parsl Parallel Scripting Library",
        "MANAGER_URL_TITLE": "Parsl",
        "MANAGER_URL": "https://parsl.readthedocs.io/en/latest/index.html",
        "MANAGER_CLIENT": "parsl_config",
        "SANITY_CHECKS": "",
        "IMPORTS": dedent("""\
            from parsl.config import Config
            from parsl.executors import HighThroughputExecutor"""
                          ),

        "MANAGER_CLIENT_BUILDER": ""
    }

    code_skeletal = dedent("""\
        parsl_config = Config(
            executors=[
                HighThroughputExecutor(
                    label='QCFractal_Compute_Executor',
                    provider={PROVIDER}(
                        {PROVIDER_OPTS}
                        scheduler_options={COMBINED_SCHEDULER_OPTS}
                        worker_init='\\n'.join(TASK_STARTUP_COMMANDS),
                        walltime="00:10:00",
                        init_blocks=1,
                        max_blocks=MAX_NODES,
                        nodes_per_block=1,        # Keep one node per block, its just easier this way
                    ),
                    # workers_per_node=MAX_TASKS_PER_NODE,
                    cores_per_worker=CORES_PER_NODE // MAX_TASKS_PER_NODE,
                    max_workers = MAX_NODES*MAX_TASKS_PER_NODE
                )

            ],
        )"""
                           )
    whitespace = ''
    for line in code_skeletal.splitlines():
        if "PROVIDER_OPTS" in line:
            whitespace = line[:len(line) - len(line.lstrip())]
            break

    # break this out down here as the "\n" causes issues
    # Also have to \\n otherwise dedent() detects it wrong
    scheduler_opts = "'{SCHEDULER_HEADER} ' + '\\n{SCHEDULER_HEADER} '.join(SCHEDULER_OPTS) + '\\n',"

    # SLURM

    slurm_imports = base_dict["IMPORTS"] + "\n" + """from parsl.providers import SlurmProvider"""

    slurm_builder = {"PROVIDER": "SlurmProvider",
                     "PROVIDER_OPTS":
                         indent(dedent("""\
                         SLURM_PARTITION,
                         exclusive=NODE_EXCLUSIVITY,"""),
                                whitespace,
                                # Don't indent first line, use whatever logic
                                predicate=lambda feed: "=" in feed),
                     "COMBINED_SCHEDULER_OPTS": scheduler_opts.format(SCHEDULER_HEADER="#SBATCH")}

    # PBS/Torque

    torque_imports = base_dict["IMPORTS"] + "\n" + """from parsl.providers import TorqueProvider"""

    torque_builder = {"PROVIDER": "TorqueProvider",
                      "PROVIDER_OPTS":
                          indent(dedent("""\
                          account=TORQUE_ACCOUNT,
                          queue=TORQUE_QUEUE,"""),
                                 whitespace,
                                 # Don't indent first line, use whatever logic
                                 predicate=lambda feed: "account" not in feed),
                      "COMBINED_SCHEDULER_OPTS": scheduler_opts.format(SCHEDULER_HEADER="#PBS")}
    # Final

    parsl_dict = {
        "slurm": {**base_dict,
                  **{"IMPORTS": slurm_imports,
                     "MANAGER_CLIENT_BUILDER": code_skeletal.format(**slurm_builder),
                     }
                  },
        "torque": {**base_dict,
                   **{"IMPORTS": torque_imports,
                      "MANAGER_CLIENT_BUILDER": code_skeletal.format(**torque_builder)}
                   },
    }

    return parsl_dict, base_dict


def parse_args():
    parser = argparse.ArgumentParser(description='A Generator for QCFractal QueueManager scripts.')
    parser.add_argument("adapter",
                        type=str,
                        help="The adapter and distribution system to build the template from")
    parser.add_argument('scheduler',
                        type=str,
                        help="The Cluster scheduler on the resource you plan to execute on (e.g. slurm, pbs/torque\n"
                             "This is capitalization agnostic, and 'torque'<->'pbs' are interchangeable")
    parser.add_argument('--test',
                        dest="test",
                        action="store_true",
                        help="Set whether or not to generate a test script or a production script")
    parser.add_argument('-o', '--output',
                        default="manager_template.py",
                        dest="output_file",
                        type=str,
                        help="Name of the output file for the template")

    args = vars(parser.parse_args())

    return args


def main(args=None):

    # Grab CLI args if not present
    if args is None:
        args = parse_args()

    adapter = args["adapter"].lower()
    valid_schedulers = ['slurm', 'torque', 'pbs', 'lsf']
    scheduler = args['scheduler'].lower()
    if scheduler == "pbs":
        scheduler = "torque"
    if scheduler not in valid_schedulers:
        raise ValueError(
            "Scheduler {} is not implemented in any adapter, must be one of {}".format(scheduler, valid_schedulers)
        )

    test = args["test"]
    test_dict = test_helper[test]

    # Handle Dask adapters
    if adapter == "dask":
        dask_dict, base_dict = dask_templates()
        if scheduler not in dask_dict:
            raise ValueError("Scheduler {} is not known by the Dask template generator. Please choose one of the "
                             "following ({}) and modify through the docs at {}".format(scheduler,
                                                                                       list(dask_dict.keys()),
                                                                                       base_dict["MANAGER_URL"]))
        template_construct = base_script.format(**test_dict, **scheduler_collections[scheduler], **dask_dict[scheduler])

    # Handle Parsl adapters
    elif adapter == "parsl":
        parsl_dict, base_dict = parsl_templates()
        if scheduler not in parsl_dict:
            raise ValueError("Scheduler {} is not known by the parsl template generator. Please choose one of the "
                             "following ({}) and modify through the docs at {}".format(scheduler,
                                                                                       list(parsl_dict.keys()),
                                                                                       base_dict["MANAGER_URL"]))
        template_construct = base_script.format(**test_dict,
                                                **scheduler_collections[scheduler],
                                                **parsl_dict[scheduler])

    # Handle Fireworks adapters
    elif adapter == "fireworks":
        raise NotImplementedError("Fireworks template generators are not yet implemented.")

    elif adapter == "executor":
        raise NotImplementedError("Executor template generators are not yet implemented.")
    else:
        raise KeyError(
            "Unknown adapter type '{}', available options: 'dask' and 'parsl'.".format(adapter))

    with open(args["output_file"], 'w') as f:
        f.write(template_construct)


if __name__ == '__main__':
    main()



"""
A command line interface to the qcfractal server.
"""

import argparse

import qcfractal

from . import cli_utils


def parse_args():
    parser = argparse.ArgumentParser(description='A CLI for the QCFractalServer.')

    manager = parser.add_argument_group('Local QueueManager Settings (optional)')
    # This option defaults to None if option not present, -1 if present, or value if provided
    manager.add_argument(
        '--local-manager',
        const=-1,
        default=None,
        action='store',
        nargs='?',
        type=int,
        help='Creates a local pool QueueManager')

    manager = parser.add_argument_group('Manager Settings')
    manager.add_argument("--heartbeat-frequency", type=int, default=300, help="The manager heartbeat frequency.")

    security = parser.add_argument_group('Security Settings')
    security.add_argument(
        "--security", type=str, default=None, choices=[None, "local"], help="The security protocol to use")
    security.add_argument(
        "--allow-read", type=bool, default=True, help="Allow read-only queries or not if security is active.")
    security.add_argument("--disable-ssl", type=bool, default=False, help="Disables SSL if present, if False a SSL cert will be created for you")
    security.add_argument("--tls-cert", type=str, default=None, help="Certificate file for TLS (in PEM format)")
    security.add_argument("--tls-key", type=str, default=None, help="Private key file for TLS (in PEM format)")

    general = parser.add_argument_group('General Settings')
    general.add_argument("database_name", type=str, help="The name of the database to use with the storage socket")
    general.add_argument("--server-name", type=str, default="QCFractal Server", help="The server name to broadcast")
    general.add_argument("--query-limit", type=int, default=1000, help="The maximum query size to server")
    general.add_argument("--log-prefix", type=str, default=None, help="The logfile prefix to use")
    general.add_argument("--database-uri", type=str, default="mongodb://localhost", help="The database URI to use")
    general.add_argument("--port", type=int, default=7777, help="The server port")
    general.add_argument("--compress-response", type=bool, default=True, help="Compress the response or not")
    general.add_argument("--config-file", type=str, default=None, help="A configuration file to use")
    general.add_argument("--start-periodics", type=bool, default=True, help="Start the periodic calls or not, always recommended unless running fractal-server behind a proxy")

    parser._action_groups.reverse()

    args = vars(parser.parse_args())
    if args["config_file"] is not None:
        data = cli_utils.read_config_file(args["config_file"])
        args = cli_utils.argparse_config_merge(parser, args, data)

    return args


def main(args=None):

    # Grab CLI args if not present
    if args is None:
        args = parse_args()

    # Handle SSL
    if args["disable_ssl"]:
        ssl_options = False
    else:
        ssl_certs = sum(args[x] is not None for x in ["tls_key", "tls_cert"])
        if ssl_certs == 0:
            ssl_options = True
        elif ssl_certs == 2:
            ssl_options = {"crt": args["tls_cert"], "key": args["tls_key"]}
        else:
            raise KeyError("Both tls-cert and tls-key must be passed in.")

    # Handle Adapters/QueueManagers
    exit_callbacks = []

    # Build an optional adapter
    if args["local_manager"]:
        ncores = args["local_manager"]
        if ncores == -1:
            ncores = None

        from concurrent.futures import ProcessPoolExecutor

        adapter = ProcessPoolExecutor(max_workers=ncores)

    else:
        adapter = None

    # Build the server itself
    server = qcfractal.FractalServer(
        name=args["server_name"],
        port=args["port"],
        compress_response=args["compress_response"],

        # Security
        security=args["security"],
        allow_read=args["allow_read"],
        ssl_options=ssl_options,

        # Database
        storage_uri=args["database_uri"],
        storage_project_name=args["database_name"],
        query_limit=args["query_limit"],

        # Log options
        logfile_prefix=args["log_prefix"],

        # Queue options
        heartbeat_frequency=args["heartbeat_frequency"],
        queue_socket=adapter)

    # Add exit callbacks
    for cb in exit_callbacks:
        server.add_exit_callback(cb[0], *cb[1], **cb[2])

    # Register closing
    cli_utils.install_signal_handlers(server.loop, server.stop)

    # Blocks until keyboard interupt
    server.start(start_periodics=args["start_periodics"])


if __name__ == '__main__':
    main()



"""
Tests the on-node procedures compute capabilities.
"""

import pytest
import requests

import qcfractal
import qcfractal.interface as ptl
from qcfractal import testing

_users = {
    "read": {
        "pw": "hello",
        "perm": ["read"]
    },
    "write": {
        "pw": "something",
        "perm": ["read", "write"]
    },
    "admin": {
        "pw": "something",
        "perm": ["read", "write", "compute", "admin"]
    }
}


@pytest.fixture(scope="module")
def sec_server(request):
    """
    Builds a server instance with the event loop running in a thread.
    """

    # Check mongo
    # testing.check_active_mongo_server()
    # storage_name = "qcf_local_server_auth_test"
    # storage_uri = "mongodb://localhost"

    testing.check_active_mongo_server()
    storage_name = "test_qcarchivedb"

    with testing.loop_in_thread() as loop:

        # Build server, manually handle IOLoop (no start/stop needed)
        server = qcfractal.FractalServer(port=testing.find_open_port(),
                                         storage_uri=testing.POSTGRES_TESTING_URI,
                                         storage_project_name=storage_name,
                                         loop=loop,
                                         security="local")

        # Clean and re-init the databse
        server.storage._clear_db(storage_name)

        # Add local users
        for k, v in _users.items():
            assert server.storage.add_user(k, _users[k]["pw"], _users[k]["perm"])

        yield server


@pytest.fixture(scope="module")
def sec_server_allow_read(sec_server):
    """
    New sec server with read allowed
    """
    yield qcfractal.FractalServer(name="qcf_server_allow_read",
                                  port=testing.find_open_port(),
                                  storage_project_name=sec_server.storage.get_project_name(),
                                  storage_uri=testing.POSTGRES_TESTING_URI,
                                  loop=sec_server.loop,
                                  security="local",
                                  allow_read=True)


### Tests the compute queue stack
def test_security_auth_decline_none(sec_server):
    with pytest.raises(IOError) as excinfo:
        client = ptl.FractalClient(sec_server)

    assert "user not found" in str(excinfo.value).lower()


def test_security_auth_bad_ssl(sec_server):
    with pytest.raises(ConnectionError) as excinfo:
        client = ptl.FractalClient.from_file({
            "address": sec_server.get_address(),
            "username": "read",
            "password": _users["write"]["pw"],
            "verify": True
        })

    assert "ssl handshake" in str(excinfo.value).lower()
    assert "verify=false" in str(excinfo.value).lower()


def test_security_auth_decline_bad_user(sec_server):
    with pytest.raises(IOError) as excinfo:
        client = ptl.FractalClient.from_file({
            "address": sec_server.get_address(),
            "username": "hello",
            "password": "something",
            "verify": False
        })

    assert "user not found" in str(excinfo.value).lower()


def test_security_auth_accept(sec_server):

    client = ptl.FractalClient(sec_server, username="write", password=_users["write"]["pw"])

    r = client.add_molecules([])
    r = client.query_molecules(id=[])


def test_security_auth_password_gen(sec_server):

    pw = sec_server.storage.add_user("autogenpw", None, ["read"])
    assert sec_server.storage.verify_user("autogenpw", pw, "read")[0]
    assert isinstance(pw, str)
    assert len(pw) > 20


def test_security_auth_overwrite(sec_server):

    user = "auth_overwrite"
    pw1 = sec_server.storage.add_user(user, None, ["read"])
    assert sec_server.storage.verify_user(user, pw1, "read")[0]

    pw2 = sec_server.storage.add_user(user, None, ["read"], overwrite=True)
    assert isinstance(pw2, str)
    assert pw1 != pw2

    assert sec_server.storage.verify_user(user, pw1, "read")[0] is False
    assert sec_server.storage.verify_user(user, pw2, "read")[0]


## Allow read tests


def test_security_auth_allow_read(sec_server_allow_read):
    client = ptl.FractalClient(sec_server_allow_read)


def test_security_auth_allow_read_block_add(sec_server_allow_read):
    client = ptl.FractalClient(sec_server_allow_read)

    with pytest.raises(IOError) as excinfo:
        client.add_molecules([ptl.Molecule.from_data("He 0 0 0")])
    assert "user not found" in str(excinfo.value).lower()



"""
QCPortal Database ODM
"""
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd
from pydantic import BaseModel

from .collection import BaseProcedureDataset
from .collection_utils import register_collection
from ..models import Molecule, ObjectId, OptimizationSpecification, QCSpecification, TorsionDriveInput
from ..models.torsiondrive import TDKeywords
from ..visualization import custom_plot


class TDRecord(BaseModel):
    """Data model for the `reactions` list in Dataset"""
    name: str
    initial_molecules: List[ObjectId]
    td_keywords: TDKeywords
    attributes: Dict[str, Union[int, float, str]]  # Might be overloaded key types
    object_map: Dict[str, ObjectId] = {}


class TorsionDriveSpecification(BaseModel):
    name: str
    description: Optional[str]
    optimization_spec: OptimizationSpecification
    qc_spec: QCSpecification


class TorsionDriveDataset(BaseProcedureDataset):
    class DataModel(BaseProcedureDataset.DataModel):

        records: Dict[str, TDRecord] = {}
        history: Set[str] = set()
        specs: Dict[str, TorsionDriveSpecification] = {}

        class Config(BaseProcedureDataset.DataModel.Config):
            pass

    def add_specification(self,
                          name: str,
                          optimization_spec: OptimizationSpecification,
                          qc_spec: QCSpecification,
                          description: str=None,
                          overwrite=False) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the specification
        optimization_spec : OptimizationSpecification
            A full optimization specification for TorsionDrive
        qc_spec : QCSpecification
            A full quantum chemistry specification for TorsionDrive
        description : str, optional
            A short text description of the specification
        overwrite : bool, optional
            Overwrite existing specification names

        """

        spec = TorsionDriveSpecification(
            name=name, optimization_spec=optimization_spec, qc_spec=qc_spec, description=description)

        return self._add_specification(name, spec, overwrite=overwrite)

    def add_entry(self,
                  name: str,
                  initial_molecules: List[Molecule],
                  dihedrals: List[Tuple[int, int, int, int]],
                  grid_spacing: List[int],
                  dihedral_ranges: Optional[List[Tuple[int, int]]]=None,
                  energy_decrease_thresh: Optional[float]=None,
                  energy_upper_limit: Optional[float]=None,
                  attributes: Dict[str, Any]=None) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the entry, will be used for the index
        initial_molecules : List[Molecule]
            The list of starting Molecules for the TorsionDrive
        dihedrals : List[Tuple[int, int, int, int]]
            A list of dihedrals to scan over
        grid_spacing : List[int]
            The grid spacing for each dihedrals
        dihedral_ranges: Optional[List[Tuple[int, int]]]
            The range limit of each dihedrals to scan, within [-180, 360]
        energy_decrease_thresh: Optional[float]
            The threshold of energy decrease to trigger activating grid points
        energy_upper_limit: Optional[float]
            The upper limit of energy relative to current global minimum to trigger activating grid points
        attributes : Dict[str, Any], optional
            Additional attributes and descriptions for the record
        """

        # Build new objects
        molecule_ids = self.client.add_molecules(initial_molecules)
        td_keywords = TDKeywords(
            dihedrals=dihedrals,
            grid_spacing=grid_spacing,
            dihedral_ranges=dihedral_ranges,
            energy_decrease_thresh=energy_decrease_thresh,
            energy_upper_limit=energy_upper_limit)

        record = TDRecord(name=name, initial_molecules=molecule_ids, td_keywords=td_keywords, attributes=attributes)

        self._add_entry(name, record)

    def compute(self, specification: str, subset: Set[str]=None, tag: Optional[str]=None,
                priority: Optional[str]=None) -> int:
        """Computes a specification for all records in the dataset.

        Parameters
        ----------
        specification : str
            The specification name.
        subset : Set[str], optional
            Computes only a subset of the dataset.
        tag : Optional[str], optional
            The queue tag to use when submitting compute requests.
        priority : Optional[str], optional
            The priority of the jobs low, medium, or high.

        Returns
        -------
        int
            The number of submitted torsiondrives
        """
        specification = specification.lower()
        spec = self.get_specification(specification)
        if subset:
            subset = set(subset)

        submitted = 0
        for rec in self.data.records.values():
            if specification in rec.object_map:
                continue

            if (subset is not None) and (rec.name not in subset):
                continue

            service = TorsionDriveInput(
                initial_molecule=rec.initial_molecules,
                keywords=rec.td_keywords,
                optimization_spec=spec.optimization_spec,
                qc_spec=spec.qc_spec)

            rec.object_map[spec.name] = self.client.add_service([service], tag=tag, priority=priority).ids[0]
            submitted += 1

        self.data.history.add(specification)
        self.save()
        return submitted

    def counts(self,
               entries: Union[str, List[str]],
               specs: Optional[Union[str, List[str]]]=None,
               count_gradients=False) -> 'DataFrame':
        """Counts the number of optimization or gradient evaluations associated with the
        TorsionDrives.

        Parameters
        ----------
        entries : Union[str, List[str]]
            The entries to query for
        specs : Optional[Union[str, List[str]]], optional
            The specifications to query for
        count_gradients : bool, optional
            If True, counts the total number of gradient calls. Warning! This can be slow for large datasets.

        Returns
        -------
        DataFrame
            The queried counts.
        """

        if isinstance(specs, str):
            specs = [specs]

        if isinstance(entries, str):
            entries = [entries]

        # Query all of the specs and make sure they are valid
        if specs is None:
            specs = list(self.df.columns)
        else:
            new_specs = []
            for spec in specs:
                new_specs.append(self.query(spec))

            # Remap names
            specs = new_specs

        # Count functions
        def count_gradient_evals(td):
            if td.status != "COMPLETE":
                return None

            total_grads = 0
            for key, optimizations in td.get_history().items():
                for opt in optimizations:
                    total_grads += len(opt.trajectory)
            return total_grads

        def count_optimizations(td):
            if td.status != "COMPLETE":
                return None
            return sum(len(v) for v in td.optimization_history.values())

        # Loop over the data and apply the count function
        ret = []
        for col in specs:
            data = self.df[col]
            if entries:
                data = data[entries]

            if count_gradients:
                cnts = data.apply(lambda td: count_gradient_evals(td))
            else:
                cnts = data.apply(lambda td: count_optimizations(td))
            ret.append(cnts)

        ret = pd.DataFrame(ret).transpose()
        ret.dropna(inplace=True, how="all")
        # ret = pd.DataFrame([ret[x].astype(int) for x in ret.columns]).transpose()
        return ret

    def visualize(self,
                  entries: Union[str, List[str]],
                  specs: Union[str, List[str]],
                  relative: bool=True,
                  units: str="kcal / mol",
                  digits: int=3,
                  use_measured_angle: bool=False,
                  return_figure: Optional[bool]=None) -> 'plotly.Figure':
        """
        Parameters
        ----------
        entries : Union[str, List[str]]
            A single or list of indices to plot.
        specs : Union[str, List[str]]
            A single or list of specifications to plot.
        relative : bool, optional
            Shows relative energy, lowest energy per scan is zero.
        units : str, optional
            The units of the plot.
        digits : int, optional
            Rounds the energies to n decimal places for display.
        use_measured_angle : bool, optional
            If True, the measured final angle instead of the constrained optimization angle.
            Can provide more accurate results if the optimization was ill-behaved,
            but pulls additional data from the server and may take longer.
        return_figure : Optional[bool], optional
            If True, return the raw plotly figure. If False, returns a hosted iPlot. If None, return a iPlot display in Jupyter notebook and a raw plotly figure in all other circumstances.

        Returns
        -------
        plotly.Figure
            The requested figure.
        """

        show_spec = True
        if isinstance(specs, str):
            specs = [specs]
            show_spec = False

        if isinstance(entries, str):
            entries = [entries]

        # Query all of the specs and make sure they are valid
        formatted_spec_names = []
        for spec in specs:
            formatted_spec_names.append(self.query(spec))

        traces = []
        ranges = []
        # Loop over specifications
        for spec in formatted_spec_names:
            # Loop over indices (groups colors by entry)
            for index in entries:

                # Plot the figure using the torsiondrives plotting function
                fig = self.df.loc[index, spec].visualize(
                    relative=relative,
                    units=units,
                    digits=digits,
                    use_measured_angle=use_measured_angle,
                    return_figure=True)

                ranges.append(fig.layout.xaxis.range)
                trace = fig.data[0]  # Pull out the underlying scatterplot

                if show_spec:
                    trace.name = f"{index}-{spec}"
                else:
                    trace.name = f"{index}"

                traces.append(trace)

        title = "TorsionDriveDataset 1-D Plot"
        if show_spec is False:
            title += f" [spec={formatted_spec_names[0]}]"

        if relative:
            ylabel = f"Relative Energy [{units}]"
        else:
            ylabel = f"Absolute Energy [{units}]"

        custom_layout = {
            "title": title,
            "yaxis": {
                "title": ylabel,
                "zeroline": True
            },
            "xaxis": {
                "title": "Dihedral Angle [degrees]",
                "zeroline": False,
                "range": [min(x[0] for x in ranges), max(x[1] for x in ranges)]
            }
        }

        return custom_plot(traces, custom_layout, return_figure=return_figure)


register_collection(TorsionDriveDataset)



from flask import current_app, g
from ..storage_sockets import storage_socket_factory


def get_socket():
    if 'socket' not in g:
        g.socket = storage_socket_factory(current_app.config['DATABASE_URI'],
                                          project_name=current_app.config['DATABASE_NAME'])

    return g.socket



"""
Contains a number of utility functions for storage sockets.
"""

import json

# Constants
_get_metadata = json.dumps({"errors": [], "n_found": 0, "success": False, "missing": [], "error_description": False})

_add_metadata = json.dumps({
    "errors": [],
    "n_inserted": 0,
    "success": False,
    "duplicates": [],
    "error_description": False,
    "validation_errors": []
})


def get_metadata_template():
    """
    Returns a copy of the metadata for database getters.
    """
    return json.loads(_get_metadata)


def add_metadata_template():
    """
    Returns a copy of the metadata for database save/updates.
    """
    return json.loads(_add_metadata)



"""
DQM Client base folder
"""

from . import collections
from . import data
from . import dict_utils
from . import models

# Add imports here
from .client import FractalClient
from .models import Molecule

# We are running inside QCPortal repo
try:
    from . import _version
    versions = _version.get_versions()
    __version__ = versions["version"]
    __git_revision__ = versions["full-revisionid"]
    _isportal = True

# We are running inside QCFractal
except ImportError:
    from ..extras import get_information
    __version__ = "inplace-{}".format(get_information('version'))
    __git_revision__ = get_information('git_revision')
    _isportal = False
    del get_information



"""
Initializer for the queue_handler folder
"""

from .adapters import build_queue_adapter
from .handlers import TaskQueueHandler, ServiceQueueHandler, QueueManagerHandler
from .managers import QueueManager



"""
A model for TorsionDrive
"""

import abc
import datetime
import json
import numpy as np
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Union

import qcelemental as qcel
from pydantic import BaseModel, constr, validator

from .common_models import DriverEnum, ObjectId, QCSpecification
from .model_utils import hash_dictionary, json_encoders, prepare_basis, recursive_normalizer
from ..visualization import scatter_plot

__all__ = ["OptimizationRecord", "ResultRecord", "OptimizationRecord"]


class RecordStatusEnum(str, Enum):
    complete = "COMPLETE"
    incomplete = "INCOMPLETE"
    running = "RUNNING"
    error = "ERROR"


class RecordBase(BaseModel, abc.ABC):
    """
    Record objects for Results and Procedures tables
    """

    # Classdata
    _hash_indices: Set[str]

    # Helper data
    client: Any = None
    cache: Dict[str, Any] = {}

    # Base identification
    id: ObjectId = None
    hash_index: Optional[str] = None
    procedure: str
    program: str
    version: int

    # Extra fields
    extras: Dict[str, Any] = {}
    stdout: Optional[ObjectId] = None
    stderr: Optional[ObjectId] = None
    error: Optional[ObjectId] = None

    # Compute status
    task_id: Optional[ObjectId] = None  # TODO: not used in SQL
    status: RecordStatusEnum = "INCOMPLETE"
    modified_on: datetime.datetime = None
    created_on: datetime.datetime = None

    # Carry-ons
    provenance: Optional[qcel.models.Provenance] = None

    class Config:
        json_encoders = json_encoders
        extra = "forbid"
        build_hash_index = True

    @validator('program')
    def check_program(cls, v):
        return v.lower()

    def __init__(self, **data):

        # Set datetime defaults if not automatically available
        data.setdefault("modified_on", datetime.datetime.utcnow())
        data.setdefault("created_on", datetime.datetime.utcnow())

        super().__init__(**data)

        # Set hash index if not present
        if self.Config.build_hash_index and (self.hash_index is None):
            self.hash_index = self.get_hash_index()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id='{self.id}' status='{self.status}')"

    def __repr__(self) -> str:
        return f"<{self}>"

### Serialization helpers

    @classmethod
    def get_hash_fields(cls):
        return cls._hash_indices | {"procedure", "program"}

    def get_hash_index(self):

        data = self.json_dict(include=self.get_hash_fields())

        return hash_dictionary(data)

    def dict(self, *args, **kwargs):
        kwargs["exclude"] = (kwargs.pop("exclude", None) or set()) | {"client", "cache"}
        # kwargs["skip_defaults"] = True
        return super().dict(*args, **kwargs)

    def json_dict(self, *args, **kwargs):
        return json.loads(self.json(*args, **kwargs))

### Checkers

    def check_client(self):
        if self.client is None:
            raise ValueError("Requested method requires a client, but client was '{}'.".format(self.client))


### KVStore Getters

    def _kvstore_getter(self, field_name):
        """
        Internal KVStore getting object
        """
        self.check_client()

        oid = self.__values__[field_name]
        if oid is None:
            return None

        if field_name not in self.cache:
            self.cache[field_name] = self.client.query_kvstore([oid])[oid]

        return self.cache[field_name]

    def get_stdout(self) -> Optional[str]:
        """Pulls the stdout from the denormalized KVStore and returns it to the user.

        Returns
        -------
        Optional[str]
            The requested stdout, none if no stdout present.
        """
        return self._kvstore_getter("stdout")

    def get_stderr(self) -> Optional[str]:
        """Pulls the stderr from the denormalized KVStore and returns it to the user.

        Returns
        -------
        Optional[str]
            The requested stderr, none if no stderr present.
        """

        return self._kvstore_getter("stderr")

    def get_error(self) -> Optional[qcel.models.ComputeError]:
        """Pulls the stderr from the denormalized KVStore and returns it to the user.

        Returns
        -------
        Optional[qcel.models.ComputeError]
            The requested compute error, none if no error present.
        """
        value = self._kvstore_getter("error")
        if value:
            return qcel.models.ComputeError(**value)
        else:
            return value


class ResultRecord(RecordBase):

    # Classdata
    _hash_indices = {"driver", "method", "basis", "molecule", "keywords", "program"}

    # Version data
    version: int = 1
    procedure: constr(strip_whitespace=True, regex="single") = "single"

    # Input data
    driver: DriverEnum
    method: str
    molecule: ObjectId
    basis: Optional[str] = None
    keywords: Optional[ObjectId] = None

    # Output data
    return_result: Union[float, List[float], Dict[str, Any]] = None
    properties: qcel.models.ResultProperties = None

    class Config(RecordBase.Config):
        build_hash_index = False

    @validator('method')
    def check_method(cls, v):
        return v.lower()

    @validator('basis')
    def check_basis(cls, v):
        return prepare_basis(v)


## QCSchema constructors

    def build_schema_input(self, molecule: 'Molecule', keywords: Optional['KeywordsSet'] = None,
                           checks: bool = True) -> 'ResultInput':
        """
        Creates a OptimizationInput schema.
        """

        if checks:
            assert self.molecule == molecule.id
            if self.keywords:
                assert self.keywords == keywords.id

        model = {"method": self.method}
        if self.basis:
            model["basis"] = self.basis

        if not self.keywords:
            keywords = {}
        else:
            keywords = keywords.values

        model = qcel.models.ResultInput(id=self.id,
                                        driver=self.driver.name,
                                        model=model,
                                        molecule=molecule,
                                        keywords=keywords,
                                        extras=self.extras)
        return model

    def consume_output(self, data: Dict[str, Any], checks: bool = True):
        assert self.method == data["model"]["method"]

        # Result specific
        self.extras = data["extras"]
        self.extras.pop("_qcfractal_tags", None)
        self.return_result = data["return_result"]
        self.properties = data["properties"]

        # Standard blocks
        self.provenance = data["provenance"]
        self.error = data["error"]
        self.stdout = data["stdout"]
        self.stderr = data["stderr"]
        self.status = "COMPLETE"


class OptimizationRecord(RecordBase):
    """
    A OptimizationRecord for all optimization procedure data.
    """

    # Classdata
    _hash_indices = {"initial_molecule", "keywords", "qc_spec"}

    # Version data
    version: int = 1
    procedure: constr(strip_whitespace=True, regex="optimization") = "optimization"
    schema_version: int = 1  # TODO: why not in Base

    # Input data
    initial_molecule: ObjectId
    qc_spec: QCSpecification
    keywords: Dict[str, Any] = {}  # TODO: defined in Base

    # Results
    energies: List[float] = None
    final_molecule: ObjectId = None
    trajectory: List[ObjectId] = None

    class Config(RecordBase.Config):
        pass

    @validator('keywords')
    def check_keywords(cls, v):
        if v is not None:
            v = recursive_normalizer(v)
        return v

## QCSchema constructors

    def build_schema_input(self,
                           initial_molecule: 'Molecule',
                           qc_keywords: Optional['KeywordsSet'] = None,
                           checks: bool = True) -> 'OptimizationInput':
        """
        Creates a OptimizationInput schema.
        """

        if checks:
            assert self.initial_molecule == initial_molecule.id
            if self.qc_spec.keywords:
                assert self.qc_spec.keywords == qc_keywords.id

        qcinput_spec = self.qc_spec.form_schema_object(keywords=qc_keywords, checks=checks)
        qcinput_spec.pop("program", None)

        model = qcel.models.OptimizationInput(id=self.id,
                                              initial_molecule=initial_molecule,
                                              keywords=self.keywords,
                                              extras=self.extras,
                                              hash_index=self.hash_index,
                                              input_specification=qcinput_spec)
        return model


## Standard function

    def get_final_energy(self):
        """The final energy of the geometry optimization.

        Returns
        -------
        float
            The optimization molecular energy.
        """
        return self.energies[-1]

    def get_trajectory(self, projection=None):
        """Returns the raw documents for each gradient evaluation in the trajectory.

        Parameters
        ----------
        client : qcportal.FractalClient
            A active client connected to a server.
        projection : None, optional
            A dictionary of the project to apply to the document

        Returns
        -------
        list of dict
            A list of results documents
        """

        return self.client.query_results(id=self.trajectory)

    def get_initial_molecule(self):
        """Returns the initial molecule

        Returns
        -------
        Molecule
            The initial molecule
        """

        ret = self.client.query_molecules(id=[self.initial_molecule])
        return ret[0]

    def get_final_molecule(self):
        """Returns the optimized molecule

        Returns
        -------
        Molecule
            The optimized molecule
        """

        ret = self.client.query_molecules(id=[self.final_molecule])
        return ret[0]

## Show functions

    def show_history(self,
                     units: str = "kcal/mol",
                     digits: int = 3,
                     relative: bool = True,
                     return_figure: Optional[bool] = None) -> 'plotly.Figure':

        cf = qcel.constants.conversion_factor("hartree", units)

        energies = np.array(self.energies)
        if relative:
            energies = energies - np.min(energies)

        trace = {
            "mode": "lines+markers",
            "x": list(range(1,
                            len(energies) + 1)),
            "y": np.around(energies * cf, digits)
        }

        if relative:
            ylabel = f"Relative Energy [{units}]"
        else:
            ylabel = f"Absolute Energy [{units}]"

        custom_layout = {
            "title": "Geometry Optimization",
            "yaxis": {
                "title": ylabel,
                "zeroline": True
            },
            "xaxis": {
                "title": "Optimization Step",
                # "zeroline": False,
                "range": [min(trace["x"]), max(trace["x"])]
            }
        }

        return scatter_plot([trace], custom_layout=custom_layout, return_figure=return_figure)



"""
    Unit tests of the SQLAlchemt to PostgreSQL interface to MongoDB

    Does not use portal or fractal server interfaces.

"""

from time import time

import pytest
import qcfractal.interface as ptl
from sqlalchemy.orm import joinedload
from qcfractal.storage_sockets.sql_models import (MoleculeORM, OptimizationProcedureORM, ResultORM,
                                                 TaskQueueORM, TorsionDriveProcedureORM, KVStoreORM,
                                                  OptimizationHistory, ServiceQueueORM, Trajectory)
from qcfractal.services.services import TorsionDriveService

from qcfractal.testing import sqlalchemy_socket_fixture as storage_socket


def session_delete_all(session, className):
    rows = session.query(className).all()
    for row in rows:
        session.delete(row)

    session.commit()
    return len(rows)

@pytest.fixture(scope='function')
def session(storage_socket):

    session = storage_socket.Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


@pytest.fixture
def molecules_H4O2(storage_socket):
    water = ptl.data.get_molecule("water_dimer_minima.psimol")
    water2 = ptl.data.get_molecule("water_dimer_stretch.psimol")

    ret = storage_socket.add_molecules([water, water2])

    yield list(ret['data'])

    r = storage_socket.del_molecules(molecule_hash=[water.get_hash(), water2.get_hash()])
    assert r == 2


@pytest.fixture
def kw_fixtures(storage_socket):
    kw1 = ptl.models.KeywordSet(**{"values": {"something": "kwfixture"}})
    ret = storage_socket.add_keywords([kw1])

    yield list(ret['data'])

    r = storage_socket.del_keywords(ret['data'][0])
    assert r == 1


def test_logs(storage_socket, session):

    assert session.query(KVStoreORM).count() == 0

    log = KVStoreORM(value="New log")
    session.add(log)
    session.commit()

    assert session.query(KVStoreORM).count() == 1

    session_delete_all(session, KVStoreORM)


def test_molecule_sql(storage_socket, session):
    """
        Test the use of the ME class MoleculeORM

        Note:
            creation of a MoleculeORM using ME is not implemented yet
            Should create a MoleculeORM using: mongoengine_socket.add_molecules
    """

    num_mol_in_db = session.query(MoleculeORM).count()
    # MoleculeORM.objects().delete()
    assert num_mol_in_db == 0

    water = ptl.data.get_molecule("water_dimer_minima.psimol")
    water2 = ptl.data.get_molecule("water_dimer_stretch.psimol")

    # Add MoleculeORM
    ret = storage_socket.add_molecules([water, water2])
    assert ret["meta"]["success"] is True
    assert ret["meta"]["n_inserted"] == 2

    ret = storage_socket.get_molecules()

    assert ret['meta']['n_found'] == 2

    # Use the ORM class
    water_mol = session.query(MoleculeORM).first()
    assert water_mol.molecular_formula == "H4O2"
    assert water_mol.molecular_charge == 0

    # print(water_mol.json_dict())
    #
    # Query with fields in the model
    result_list = session.query(MoleculeORM).filter_by(molecular_formula="H4O2").all()
    assert len(result_list) == 2
    assert result_list[0].molecular_multiplicity == 1

    # Query with fields NOT in the model. works too!
    result_list = session.query(MoleculeORM).filter_by(molecular_charge=0).all()
    assert len(result_list) == 2

    # get unique by hash and formula
    one_mol = session.query(MoleculeORM).filter_by(molecule_hash=water_mol.molecule_hash,
                                                   molecular_formula=water_mol.molecular_formula)
    assert len(one_mol.all()) == 1

    # Clean up
    storage_socket.del_molecules(molecule_hash=[water.get_hash(), water2.get_hash()])

def test_services(storage_socket, session):

    assert session.query(OptimizationProcedureORM).count() == 0

    proc_data = {
        "initial_molecule":None,
        "keywords": None,
        "program": "p7",
        "qc_spec": {
            "basis": "b1",
            "program": "p1",
            "method": "m1",
            "driver": "energy"
        },
        "status": "COMPLETE",
    }

    service_data = {
        "tag": "tag1 tag2",
        "hash_index" : "123",
        "status": "COMPLETE",

        # extra fields
        "torsiondrive_state": {},

        "dihedral_template": "1",
        "optimization_template": "2",
        "molecule_template": "",
        "logger": None,
        "storage_socket": storage_socket,
        "task_priority": 0
    }


    procedure = OptimizationProcedureORM(**proc_data)
    session.add(procedure)
    session.commit()
    assert procedure.id

    service_pydantic = TorsionDriveService(**service_data)

    doc = ServiceQueueORM(**service_pydantic.json_dict(include=ServiceQueueORM.__dict__.keys()))
    doc.extra = service_pydantic.json_dict(exclude=ServiceQueueORM.__dict__.keys())
    doc.procedure_id = procedure.id
    session.add(doc)
    session.commit()

    session.delete(doc)
    session.delete(procedure)
    session.commit()

    assert session.query(ServiceQueueORM).count() == 0


def test_results_sql(storage_socket, session, molecules_H4O2, kw_fixtures):
    """
        Handling results throught the ME classes
    """


    assert session.query(ResultORM).count() == 0

    assert len(molecules_H4O2) == 2
    assert len(kw_fixtures) == 1

    page1 = {
        "molecule": molecules_H4O2[0],
        "method": "m1",
        "basis": "b1",
        "keywords": None,
        "program": "p1",
        "driver": "energy",
        "status": "COMPLETE",
    }

    page2 = {
        "molecule": molecules_H4O2[1],
        "method": "m2",
        "basis": "b1",
        "keywords": kw_fixtures[0],
        "program": "p1",
        "driver": "energy",
        "status": "COMPLETE",
    }

    result = ResultORM(**page1)
    session.add(result)
    session.commit()

    # IMPORTANT: To be able to access lazy loading children use joinedload
    ret = session.query(ResultORM).options(joinedload('molecule_obj')).filter_by(method='m1').first()
    assert ret.molecule_obj.molecular_formula == 'H4O2'
    # Accessing the keywords_obj will issue a DB access
    assert ret.keywords_obj == None

    result2 = ResultORM(**page2)
    session.add(result2)
    session.commit()
    ret = session.query(ResultORM).options(joinedload('molecule_obj')).filter_by(method='m2').first()
    assert ret.molecule_obj.molecular_formula == 'H4O2'
    assert ret.method == "m2"

    # clean up
    session_delete_all(session, ResultORM)


def test_optimization_procedure(storage_socket, session, molecules_H4O2):
    """
        Optimization procedure
    """

    assert session.query(OptimizationProcedureORM).count() == 0
    # assert Keywords.objects().count() == 0

    data1 = {
        "initial_molecule": molecules_H4O2[0],
        "keywords": None,
        "program": "p7",
        "qc_spec": {
            "basis": "b1",
            "program": "p1",
            "method": "m1",
            "driver": "energy"
        },
        "status": "COMPLETE",
    }

    result1 = {
        "molecule": molecules_H4O2[0],
        "method": "m1",
        "basis": "b1",
        "keywords": None,
        "program": "p1",
        "driver": "energy",
        "status": "COMPLETE",
    }

    procedure = OptimizationProcedureORM(**data1)
    session.add(procedure)
    session.commit()
    proc = session.query(OptimizationProcedureORM).options(
                         joinedload('initial_molecule_obj')).first()
    assert proc.initial_molecule_obj.molecular_formula == 'H4O2'
    assert proc.procedure == 'optimization'

    # add a trajectory result
    result = ResultORM(**result1)
    session.add(result)
    session.commit()
    assert result.id

    # link result to the trajectory
    proc.trajectory_obj = [Trajectory(opt_id=proc.id, result_id=result.id)]
    session.commit()
    proc = session.query(OptimizationProcedureORM).options(
                         joinedload('trajectory_obj')).first()
    assert proc.trajectory_obj

    # clean up
    session_delete_all(session, ResultORM)
    session_delete_all(session, OptimizationProcedureORM)


def test_torsiondrive_procedure(storage_socket, session):
    """
        Torsiondrive procedure
    """

    assert session.query(TorsionDriveProcedureORM).count() == 0
    # assert Keywords.objects().count() == 0

    # molecules = MoleculeORM.objects(molecular_formula='H4O2')
    # assert molecules.count() == 2

    data1 = {
        # "molecule": molecules[0],
        "keywords": None,
        "program": "p9",
        "qc_spec": {
            "basis": "b1",
            "program": "p1",
            "method": "m1",
            "driver": "energy"
        },
        "status": "COMPLETE",
    }

    torj_proc = TorsionDriveProcedureORM(**data1)
    session.add(torj_proc)
    session.commit()

    # Add optimization_history

    opt_proc = OptimizationProcedureORM(**data1)
    opt_proc2 = OptimizationProcedureORM(**data1)
    session.add_all([opt_proc, opt_proc2])
    session.commit()
    assert opt_proc.id

    opt_hist = OptimizationHistory(torsion_id=torj_proc.id, opt_id=opt_proc.id, key='20')
    opt_hist2 = OptimizationHistory(torsion_id=torj_proc.id, opt_id=opt_proc2.id, key='20')
    torj_proc.optimization_history_obj = [opt_hist, opt_hist2]
    session.commit()
    torj_proc = session.query(TorsionDriveProcedureORM).options(joinedload('optimization_history_obj')).first()
    assert torj_proc.optimization_history == {'20': [str(opt_proc.id), str(opt_proc2.id)]}

    # clean up
    session_delete_all(session, OptimizationProcedureORM)
    session_delete_all(session, TorsionDriveProcedureORM)


def test_add_task_queue(storage_socket, session, molecules_H4O2):
    """
        Simple test of adding a task using the SQL classes
        in QCFractal, tasks should be added using storage_socket
    """

    assert session.query(TaskQueueORM).count() == 0
    # TaskQueueORM.objects().delete()

    page1 = {
        "molecule": molecules_H4O2[0],
        "method": "m1",
        "basis": "b1",
        "keywords": None,
        "program": "p1",
        "driver": "energy",
    }
    # add a task that reference results
    result = ResultORM(**page1)
    session.add(result)
    session.commit()

    task = TaskQueueORM(base_result_obj=result)
    session.add(task)
    session.commit()

    ret = session.query(TaskQueueORM)
    assert  ret.count() == 1

    task = ret.first()
    assert task.status == 'WAITING'
    assert task.base_result_obj.status == 'INCOMPLETE'

    # cleanup
    session_delete_all(session, TaskQueueORM)
    session_delete_all(session, ResultORM)



def test_results_pagination(storage_socket, session, molecules_H4O2, kw_fixtures):
    """
        Test results pagination
    """

    assert session.query(ResultORM).count() == 0

    result_template = {
        "molecule": molecules_H4O2[0],
        "method": "m1",
        "basis": "b1",
        "keywords": kw_fixtures[0],
        "program": "p1",
        "driver": "energy",
    }


    # Save ~ 1 msec/doc in ME, 0.5 msec/doc in SQL
    # ------------------------------------------
    t1 = time()

    total_results = 1000
    first_half = int(total_results / 2)
    limit = 100
    skip = 50

    for i in range(first_half):
        result_template['basis'] = str(i)
        r = ResultORM(**result_template)
        session.add(r)

    result_template['method'] = 'm2'
    for i in range(first_half, total_results):
        result_template['basis'] = str(i)
        r = ResultORM(**result_template)
        session.add(r)

    session.commit()  # must commit outside the loop, 10 times faster

    total_time = (time() - t1) * 1000 / total_results
    print('Inserted {} results in {:.2f} msec / doc'.format(total_results, total_time))

    # query (~ 0.13 msec/doc) in ME, and ~0.02 msec/doc in SQL
    # ----------------------------------------
    t1 = time()

    ret1 = session.query(ResultORM).filter_by(method='m1')
    ret2 = session.query(ResultORM).filter_by(method='m2') .limit(limit) #.offset(skip)

    data1 = [d.to_dict() for d in ret1]
    data2 = [d.to_dict() for d in ret2]

    # count is total, but actual data size is the limit
    assert ret1.count() == first_half
    assert len(data1) == first_half

    # assert ret2.count() == total_results - first_half
    # assert len(ret2) == limit
    # assert len(data2) == limit
    #
    # assert int(data2[0]['basis']) == first_half + skip
    #
    # # get the last page when with fewer than limit are remaining
    # ret = session.query(ResultORM).filter_by(method='m1').limit(limit).offset(int(first_half - limit / 2))
    # assert len(ret) == limit / 2

    total_time = (time() - t1) * 1000 / total_results
    print('Query {} results in {:.3f} msec /doc'.format(total_results, total_time))

    # cleanup
    session_delete_all(session, ResultORM)




"""
A model for GridOptimization
"""
import copy
import json
from typing import Any, Dict, List, Tuple, Union

from pydantic import BaseModel, constr, validator

from .common_models import Molecule, ObjectId, OptimizationSpecification, QCSpecification
from .model_utils import json_encoders, recursive_normalizer
from .records import RecordBase

__all__ = ["GridOptimizationInput", "GridOptimizationRecord"]


class ScanDimension(BaseModel):
    """
    A dimension to scan over
    """
    type: str
    indices: List[int]
    steps: List[float]
    step_type: str

    class Config:
        extra = "forbid"
        allow_mutation = False

    @validator('type')
    def check_type(cls, v):
        v = v.lower()
        possibilities = {"distance", "angle", "dihedral"}
        if v not in possibilities:
            raise TypeError("Type '{}' found, can only be one of {}.".format(v, possibilities))

        return v

    @validator('indices', whole=True)
    def check_indices(cls, v, values, **kwargs):
        sizes = {"distance": 2, "angle": 3, "dihedral": 4}
        if sizes[values["type"]] != len(v):
            raise ValueError("ScanDimension of type {} must have {} values, found {}.".format(
                values["type"], sizes[values["type"]], len(v)))

        return v

    @validator('steps', whole=True)
    def check_steps(cls, v):
        if not (all(x < y for x, y in zip(v, v[1:])) or all(x > y for x, y in zip(v, v[1:]))):
            raise ValueError("Steps are not strictly monotonically increasing or decreasing.")

        v = recursive_normalizer(v)

        return v

    @validator('step_type')
    def check_step_type(cls, v):
        v = v.lower()
        if v not in ["absolute", "relative"]:
            raise KeyError("Keyword 'step_type' must either be absolute or relative.")

        return v


class GOKeywords(BaseModel):
    """
    GridOptimizationRecord options
    """
    scans: List[ScanDimension]
    preoptimization: bool = True

    class Config:
        extra = "forbid"
        allow_mutation = False


_gridopt_constr = constr(strip_whitespace=True, regex="gridoptimization")
_qcfractal_constr = constr(strip_whitespace=True, regex="qcfractal")


class GridOptimizationInput(BaseModel):
    """
    A GridOptimizationRecord Input base class
    """

    program: _qcfractal_constr = "qcfractal"
    procedure: _gridopt_constr = "gridoptimization"
    initial_molecule: Union[ObjectId, Molecule]
    keywords: GOKeywords
    optimization_spec: OptimizationSpecification
    qc_spec: QCSpecification

    class Config:
        allow_mutation = False
        json_encoders = json_encoders


class GridOptimizationRecord(RecordBase):
    """
    A interface to the raw JSON data of a GridOptimizationRecord torsion scan run.
    """

    # Classdata
    _hash_indices = {"initial_molecule", "keywords", "optimization_meta", "qc_spec"}

    # Version data
    version: int = 1
    procedure: _gridopt_constr = "gridoptimization"
    program: _qcfractal_constr = "qcfractal"

    # Input data
    initial_molecule: ObjectId
    keywords: GOKeywords
    optimization_spec: OptimizationSpecification
    qc_spec: QCSpecification

    # Output data
    starting_molecule: ObjectId
    final_energy_dict: Dict[str, float]
    grid_optimizations: Dict[str, ObjectId]
    starting_grid: tuple

    class Config(RecordBase.Config):
        pass

## Utility

    def _organize_return(self, data: Dict[str, Any], key: Union[int, str, None]) -> Dict[str, Any]:

        if key is None:
            return {self.deserialize_key(k): copy.deepcopy(v) for k, v in data.items()}
        else:
            key = self.serialize_key(key)

        return copy.deepcopy(data[key])

    def serialize_key(self, key):
        if isinstance(key, (int, float)):
            key = (int(key), )

        return json.dumps(key)

    def deserialize_key(self, key):

        return tuple(json.loads(key))

    def get_scan_value(self, key: Union[str, Tuple]) -> Tuple:
        """
        Obtains the scan parameters at a given grid point.
        """
        if isinstance(key, str):
            key = self.deserialize_key(key)

        ret = []
        for n, idx in enumerate(key):
            ret.append(self.keywords.scans[n].steps[idx])

        return tuple(ret)

    def get_scan_dimensions(self) -> Tuple:
        """
        Returns the overall dimensions of the scan.
        """
        ret = []
        for scan in self.keywords.scans:
            ret.append(len(scan.steps))

        return tuple(ret)

## Query

    def get_final_energies(self, key: Union[int, str, None]=None) -> Dict[str, Any]:
        """
        Provides the final optimized energies at each grid point.

        Parameters
        ----------
        key : Union[int, str, None], optional
            Specifies a single entry to pull from.

        Returns
        -------
        energy : Dict[str, Any]
            Returns energies at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> grid_optimization_record.get_final_energies()
        {(-90,): -148.7641654446243, (180,): -148.76501336993732, (0,): -148.75056290106735, (90,): -148.7641654446148}

        >>> grid_optimization_record.get_final_energies((-90,))
        -148.7641654446243

        """

        return self._organize_return(self.final_energy_dict, key)


    def get_final_molecules(self, key: Union[int, str, None]=None) -> Dict[str, Any]:
        """
        Provides the final optimized molecules at each grid point.

        Parameters
        ----------
        key : Union[int, str, None], optional
            Specifies a single entry to pull from.


        Returns
        -------
        final_molecules : Dict[str, Any]
            Returns energies at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------

        >>> mols = grid_optimization_record.get_final_molecules()
        >>> type(mols[(-90, )])
        qcelemental.models.molecule.Molecule

        >>> type(grid_optimization_record.get_final_molecules((-90,)))
        qcelemental.models.molecule.Molecule

        """

        if "final_molecules" not in self.cache:
            ret = {}
            for k, task_id in self.grid_optimizations.items():
                task = self.client.query_procedures(id=task_id)[0]
                ret[k] = task.get_final_molecule()
            self.cache["final_molecules"] = ret

        data = self.cache["final_molecules"]
        return self._organize_return(data, key)


    def get_final_results(self, key: Union[int, Tuple[int, ...], str]=None) -> Dict[str, Any]:
        """Returns the final opt gradient result records at each grid point.

        Parameters
        ----------
        key : Union[int, str, None], optional
            Specifies a single entry to pull from.

        Returns
        -------
        final_results : Dict[str, qcfractal.interface.models.records.ResultRecord]
            Returns ResultRecord at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> mols = grid_optimization_record.get_final_results()
        >>> type(mols[(-90, )])
        qcfractal.interface.models.records.ResultRecord

        >>> type(grid_optimization_record.get_final_results((-90,)))
        qcfractal.interface.models.records.ResultRecord

        """

        if "final_results" not in self.cache:
            map_id_key = {}
            ret = {}
            for k, task_id in self.grid_optimizations.items():
                task = self.client.query_procedures(id=task_id)[0]
                if len(task.trajectory) > 0:
                    final_grad_record_id = task.trajectory[-1]
                    # store the id -> grid id mapping
                    map_id_key[final_grad_record_id] = k
            # combine the ids into one query
            query_result_ids = list(map_id_key.keys())
            # run the query on this batch
            for grad_result_record in self.client.query_results(id=query_result_ids):
                k = map_id_key[grad_result_record.id]
                ret[k] = grad_result_record

            self.cache["final_results"] = ret

        data = self.cache["final_results"]

        return self._organize_return(data, key)
